from __future__ import annotations

import datetime
import decimal
import uuid
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, Self, cast
from uuid import UUID

import graphene
import graphene_federation
import jwt
import sqlalchemy as sa
from dateutil.parser import parse as dtparse
from dateutil.tz import tzutc
from graphene.types.datetime import DateTime as GQLDateTime
from graphql import Undefined, UndefinedType
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import joinedload, selectinload

from ai.backend.common.data.endpoint.types import EndpointStatus
from ai.backend.common.exception import InvalidAPIParameters
from ai.backend.common.types import (
    MODEL_SERVICE_RUNTIME_PROFILES,
    AutoScalingMetricComparator,
    AutoScalingMetricSource,
    ClusterMode,
    EndpointId,
    MountPermission,
    MountTypes,
    ResourceSlot,
    RuleId,
    RuntimeVariant,
)
from ai.backend.manager.data.deployment.types import RouteStatus
from ai.backend.manager.data.model_serving.creator import EndpointAutoScalingRuleCreator
from ai.backend.manager.data.model_serving.modifier import (
    ExtraMount,
    ImageRef,
)
from ai.backend.manager.data.model_serving.types import (
    EndpointAutoScalingRuleData,
    EndpointData,
)
from ai.backend.manager.errors.common import (
    GenericForbidden,
    ObjectNotFound,
)
from ai.backend.manager.errors.service import (
    EndpointNotFound,
    EndpointTokenNotFound,
)
from ai.backend.manager.models.endpoint import (
    EndpointAutoScalingRuleRow,
    EndpointLifecycle,
    EndpointRow,
    EndpointTokenRow,
)
from ai.backend.manager.models.image import ImageRow
from ai.backend.manager.models.minilang import FieldSpecItem, OrderSpecItem
from ai.backend.manager.models.minilang.ordering import QueryOrderParser
from ai.backend.manager.models.minilang.queryfilter import QueryFilterParser
from ai.backend.manager.models.user import UserRole, UserRow
from ai.backend.manager.models.vfolder import VFolderRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.model_serving.updaters import (
    EndpointAutoScalingRuleUpdaterSpec,
    EndpointUpdaterSpec,
)
from ai.backend.manager.services.model_serving.actions.create_auto_scaling_rule import (
    CreateEndpointAutoScalingRuleAction,
)
from ai.backend.manager.services.model_serving.actions.delete_auto_scaling_rule import (
    DeleteEndpointAutoScalingRuleAction,
)
from ai.backend.manager.services.model_serving.actions.modify_auto_scaling_rule import (
    ModifyEndpointAutoScalingRuleAction,
)
from ai.backend.manager.services.model_serving.actions.modify_endpoint import ModifyEndpointAction
from ai.backend.manager.types import OptionalState, TriState

from .base import (
    FilterExprArg,
    ImageRefType,
    InferenceSessionError,
    Item,
    OrderExprArg,
    PaginatedList,
    generate_sql_info_for_gql_connection,
)
from .gql_relay import AsyncNode, Connection, ConnectionResolverResult
from .image import ImageNode
from .routing import Routing
from .vfolder import VirtualFolderNode

if TYPE_CHECKING:
    from .schema import GraphQueryContext


AutoScalingMetricSourceGQLEnum = graphene.Enum.from_enum(
    AutoScalingMetricSource,
    description="The source type to fetch metrics. Added in 25.1.0.",
)
AutoScalingMetricComparatorGQLEnum = graphene.Enum.from_enum(
    AutoScalingMetricComparator,
    description="The comparator used to compare the metric value with the threshold. Added in 25.1.0.",
)


@graphene_federation.key("id")
class EndpointAutoScalingRuleNode(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (AsyncNode,)
        description = "Added in 25.1.0."

    row_id = graphene.UUID(required=True)

    metric_source = graphene.Field(
        AutoScalingMetricSourceGQLEnum,
        required=True,
    )
    metric_name = graphene.String(required=True)
    threshold = graphene.String(required=True)
    comparator = graphene.Field(
        AutoScalingMetricComparatorGQLEnum,
        required=True,
    )
    step_size = graphene.Int(required=True)
    cooldown_seconds = graphene.Int(required=True)

    min_replicas = graphene.Int()
    max_replicas = graphene.Int()

    created_at = GQLDateTime(required=True)
    last_triggered_at = GQLDateTime()

    endpoint = graphene.UUID(required=True)

    _queryfilter_fieldspec: Mapping[str, FieldSpecItem] = {
        "id": ("id", None),
        "metric_source": ("metric_source", None),
        "metric_name": ("metric_name", None),
        "threshold": ("threshold", None),
        "comparator": ("comparator", None),
        "step_size": ("step_size", None),
        "cooldown_seconds": ("cooldown_seconds", None),
        "created_at": ("created_at", dtparse),
        "last_triggered_at": ("last_triggered_at", dtparse),
        "endpoint": ("endpoint", None),
    }

    _queryorder_colmap: Mapping[str, OrderSpecItem] = {
        "id": ("id", None),
        "metric_source": ("metric_source", None),
        "metric_name": ("metric_name", None),
        "threshold": ("threshold", None),
        "comparator": ("comparator", None),
        "step_size": ("step_size", None),
        "cooldown_seconds": ("cooldown_seconds", None),
        "created_at": ("created_at", None),
        "last_triggered_at": ("last_triggered_at", None),
        "endpoint": ("endpoint", None),
    }

    @classmethod
    def from_dto(cls, dto: EndpointAutoScalingRuleData) -> Self:
        return cls(
            id=dto.id,
            row_id=dto.id,
            metric_source=dto.metric_source,
            metric_name=dto.metric_name,
            threshold=dto.threshold,
            comparator=dto.comparator,
            step_size=dto.step_size,
            cooldown_seconds=dto.cooldown_seconds,
            min_replicas=dto.min_replicas,
            max_replicas=dto.max_replicas,
            created_at=dto.created_at,
            last_triggered_at=dto.last_triggered_at,
            endpoint=dto.endpoint,
        )

    @classmethod
    def from_row(cls, graph_ctx: GraphQueryContext, row: EndpointAutoScalingRuleRow) -> Self:
        return cls(
            id=row.id,
            row_id=row.id,
            metric_source=row.metric_source,
            metric_name=row.metric_name,
            threshold=row.threshold,
            comparator=row.comparator,
            step_size=row.step_size,
            cooldown_seconds=row.cooldown_seconds,
            min_replicas=row.min_replicas,
            max_replicas=row.max_replicas,
            created_at=row.created_at,
            last_triggered_at=row.last_triggered_at,
            endpoint=row.endpoint,
        )

    @classmethod
    async def get_node(cls, info: graphene.ResolveInfo, rule_id: str) -> Self:
        graph_ctx: GraphQueryContext = info.context

        _, raw_rule_id = AsyncNode.resolve_global_id(info, rule_id)
        if not raw_rule_id:
            raw_rule_id = rule_id
        try:
            _rule_id = RuleId(UUID(raw_rule_id))
        except ValueError as e:
            raise ObjectNotFound(object_name="Endpoint Autoscaling Rule") from e

        async with graph_ctx.db.begin_readonly_session() as db_session:
            rule_row = await EndpointAutoScalingRuleRow.get(
                db_session, _rule_id, load_endpoint=True
            )
            match graph_ctx.user["role"]:
                case UserRole.SUPERADMIN:
                    pass
                case UserRole.ADMIN:
                    if rule_row.endpoint_row.domain != graph_ctx.user["domain_name"]:
                        raise GenericForbidden
                case UserRole.USER:
                    if rule_row.endpoint_row.created_user != graph_ctx.user["uuid"]:
                        raise GenericForbidden

            return cls.from_row(graph_ctx, rule_row)

    @classmethod
    async def get_connection(
        cls,
        info: graphene.ResolveInfo,
        endpoint: str,
        *,
        filter_expr: str | None = None,
        order_expr: str | None = None,
        offset: int | None = None,
        after: str | None = None,
        first: int | None = None,
        before: str | None = None,
        last: int | None = None,
    ) -> ConnectionResolverResult[Self]:
        graph_ctx: GraphQueryContext = info.context
        _filter_arg = (
            FilterExprArg(filter_expr, QueryFilterParser(cls._queryfilter_fieldspec))
            if filter_expr is not None
            else None
        )
        _order_expr = (
            OrderExprArg(order_expr, QueryOrderParser(cls._queryorder_colmap))
            if order_expr is not None
            else None
        )
        (
            query,
            cnt_query,
            _,
            cursor,
            pagination_order,
            page_size,
        ) = generate_sql_info_for_gql_connection(
            info,
            EndpointAutoScalingRuleRow,
            EndpointAutoScalingRuleRow.id,
            _filter_arg,
            _order_expr,
            offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )

        async with graph_ctx.db.begin_readonly_session() as db_session:
            _, raw_endpoint_id = AsyncNode.resolve_global_id(info, endpoint)
            if not raw_endpoint_id:
                raw_endpoint_id = endpoint
            try:
                _endpoint_id = EndpointId(UUID(raw_endpoint_id))
            except ValueError as e:
                raise ObjectNotFound(object_name="Endpoint") from e
            try:
                row = await EndpointRow.get(db_session, _endpoint_id)
            except NoResultFound as e:
                raise ObjectNotFound(object_name="Endpoint") from e

            match graph_ctx.user["role"]:
                case UserRole.SUPERADMIN:
                    pass
                case UserRole.ADMIN:
                    if row.domain != graph_ctx.user["domain_name"]:
                        raise GenericForbidden
                case UserRole.USER:
                    if row.created_user != graph_ctx.user["uuid"]:
                        raise GenericForbidden

            query = query.filter(EndpointAutoScalingRuleRow.endpoint == _endpoint_id)
            group_rows = (await db_session.scalars(query)).all()
            result = [cls.from_row(graph_ctx, row) for row in group_rows]
            total_cnt = await db_session.scalar(cnt_query)
            return ConnectionResolverResult(result, cursor, pagination_order, page_size, total_cnt)


class EndpointAutoScalingRuleConnection(Connection):
    class Meta:
        node = EndpointAutoScalingRuleNode
        description = "Added in 25.1.0."


class EndpointAutoScalingRuleInput(graphene.InputObjectType):  # type: ignore[misc]
    class Meta:
        description = "Added in 25.1.0."

    metric_source = graphene.Field(
        AutoScalingMetricSourceGQLEnum,
        required=True,
    )
    metric_name = graphene.String(required=True)
    threshold = graphene.String(required=True)
    comparator = graphene.Field(
        AutoScalingMetricComparatorGQLEnum,
        required=True,
    )
    step_size = graphene.Int(required=True)
    cooldown_seconds = graphene.Int(required=True)
    min_replicas = graphene.Int()
    max_replicas = graphene.Int()

    def to_action(self, endpoint_id: EndpointId) -> CreateEndpointAutoScalingRuleAction:
        return CreateEndpointAutoScalingRuleAction(
            endpoint_id=endpoint_id,
            creator=EndpointAutoScalingRuleCreator(
                metric_source=AutoScalingMetricSource(self.metric_source),
                metric_name=self.metric_name,
                threshold=self.threshold,
                comparator=AutoScalingMetricComparator(self.comparator),
                step_size=self.step_size,
                cooldown_seconds=self.cooldown_seconds,
                min_replicas=self.min_replicas if self.min_replicas is not Undefined else None,
                max_replicas=self.max_replicas if self.max_replicas is not Undefined else None,
            ),
        )


class ModifyEndpointAutoScalingRuleInput(graphene.InputObjectType):  # type: ignore[misc]
    class Meta:
        description = "Added in 25.1.0."

    metric_source = graphene.Field(
        AutoScalingMetricSourceGQLEnum,
        default_value=Undefined,
    )
    metric_name = graphene.String()
    threshold = graphene.String()
    comparator = graphene.Field(
        AutoScalingMetricComparatorGQLEnum,
        default_value=Undefined,
    )
    step_size = graphene.Int()
    cooldown_seconds = graphene.Int()
    min_replicas = graphene.Int()
    max_replicas = graphene.Int()

    def to_action(self, id: RuleId) -> ModifyEndpointAutoScalingRuleAction:
        def convert_to_decimal(
            value: str | None | UndefinedType,
        ) -> decimal.Decimal | UndefinedType:
            if isinstance(value, UndefinedType):
                return value
            if value is None:
                raise InvalidAPIParameters("Threshold cannot be None")

            try:
                return decimal.Decimal(value)
            except decimal.InvalidOperation as e:
                raise InvalidAPIParameters(f"Cannot convert {value} to Decimal") from e

        spec = EndpointAutoScalingRuleUpdaterSpec(
            metric_source=OptionalState.from_graphql(
                AutoScalingMetricSource(self.metric_source)
                if self.metric_source is not Undefined
                else Undefined,
            ),
            metric_name=OptionalState.from_graphql(
                self.metric_name,
            ),
            threshold=OptionalState.from_graphql(
                convert_to_decimal(self.threshold),
            ),
            comparator=OptionalState.from_graphql(
                AutoScalingMetricComparator(self.comparator)
                if self.comparator is not Undefined
                else Undefined,
            ),
            step_size=OptionalState.from_graphql(
                self.step_size,
            ),
            cooldown_seconds=OptionalState.from_graphql(
                self.cooldown_seconds,
            ),
            min_replicas=TriState.from_graphql(
                self.min_replicas,
            ),
            max_replicas=TriState.from_graphql(
                self.max_replicas,
            ),
        )
        return ModifyEndpointAutoScalingRuleAction(
            id=id,
            updater=Updater(spec=spec, pk_value=id),
        )


class CreateEndpointAutoScalingRuleNode(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.USER, UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        endpoint = graphene.String(required=True)
        props = EndpointAutoScalingRuleInput(required=True)

    class Meta:
        description = "Added in 25.1.0."

    ok = graphene.Boolean()
    msg = graphene.String()
    rule = graphene.Field(lambda: EndpointAutoScalingRuleNode, required=False)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        endpoint: str,
        props: EndpointAutoScalingRuleInput,
    ) -> Self:
        graph_ctx: GraphQueryContext = info.context
        _, raw_endpoint_id = AsyncNode.resolve_global_id(info, endpoint)
        if not raw_endpoint_id:
            raw_endpoint_id = endpoint
        try:
            _endpoint_id = EndpointId(UUID(raw_endpoint_id))
        except ValueError as e:
            raise ObjectNotFound(object_name="Endpoint") from e

        action = props.to_action(
            endpoint_id=_endpoint_id,
        )

        result = await graph_ctx.processors.model_serving_auto_scaling.create_endpoint_auto_scaling_rule.wait_for_complete(
            action
        )

        return cls(
            ok=result.success,
            msg="Auto scaling rule created"
            if result.success
            else "Failed to create auto scaling rule",
            rule=EndpointAutoScalingRuleNode.from_dto(result.data) if result.data else None,
        )


class ModifyEndpointAutoScalingRuleNode(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.USER, UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        id = graphene.String(required=True)
        props = ModifyEndpointAutoScalingRuleInput(required=True)

    class Meta:
        description = "Added in 25.1.0."

    ok = graphene.Boolean()
    msg = graphene.String()
    rule = graphene.Field(lambda: EndpointAutoScalingRuleNode, required=False)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        id: str,
        props: ModifyEndpointAutoScalingRuleInput,
    ) -> Self:
        _, rule_id = AsyncNode.resolve_global_id(info, id)
        if not rule_id:
            rule_id = id
        try:
            _rule_id = RuleId(UUID(rule_id))
        except ValueError as e:
            raise ObjectNotFound(object_name="Endpoint Autoscaling Rule") from e
        graph_ctx: GraphQueryContext = info.context

        action = props.to_action(
            id=_rule_id,
        )

        result = await graph_ctx.processors.model_serving_auto_scaling.modify_endpoint_auto_scaling_rule.wait_for_complete(
            action
        )

        return cls(
            ok=result.success,
            msg="Auto scaling rule updated"
            if result.success
            else "Failed to update auto scaling rule",
            rule=EndpointAutoScalingRuleNode.from_dto(result.data) if result.data else None,
        )


class DeleteEndpointAutoScalingRuleNode(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.USER, UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        id = graphene.String(required=True)

    class Meta:
        description = "Added in 25.1.0."

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        id: str,
    ) -> Self:
        _, rule_id = AsyncNode.resolve_global_id(info, id)
        try:
            _rule_id = RuleId(UUID(rule_id))
        except ValueError as e:
            raise ObjectNotFound(object_name="Endpoint Autoscaling Rule") from e

        graph_ctx: GraphQueryContext = info.context

        action = DeleteEndpointAutoScalingRuleAction(
            id=_rule_id,
        )

        result = await graph_ctx.processors.model_serving_auto_scaling.delete_endpoint_auto_scaling_rule.wait_for_complete(
            action
        )

        return cls(
            ok=result.success,
            msg="Auto scaling rule deleted"
            if result.success
            else "Failed to delete auto scaling rule",
        )


class RuntimeVariantInfo(graphene.ObjectType):  # type: ignore[misc]
    """Added in 24.03.5."""

    name = graphene.String()
    human_readable_name = graphene.String()

    @classmethod
    def from_enum(cls, enum: RuntimeVariant) -> Self:
        return cls(name=enum.value, human_readable_name=MODEL_SERVICE_RUNTIME_PROFILES[enum].name)


class Endpoint(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (Item,)

    endpoint_id = graphene.UUID()
    image = graphene.String(deprecation_reason="Deprecated since 23.09.9. use `image_object`")
    image_object = graphene.Field(ImageNode, description="Added in 23.09.9.")
    domain = graphene.String()
    project = graphene.String()
    resource_group = graphene.String()
    resource_slots = graphene.JSONString()
    url = graphene.String()
    model = graphene.UUID()
    model_definition_path = graphene.String(description="Added in 24.03.4.")
    model_vfolder = VirtualFolderNode()
    model_mount_destiation = graphene.String(
        deprecation_reason="Deprecated since 24.03.4; use `model_mount_destination` instead"
    )
    model_mount_destination = graphene.String(description="Added in 24.03.4.")
    extra_mounts = graphene.List(lambda: VirtualFolderNode, description="Added in 24.03.4.")
    created_user = graphene.UUID(
        deprecation_reason="Deprecated since 23.09.8. use `created_user_id`"
    )
    created_user_email = graphene.String(description="Added in 23.09.8.")
    created_user_id = graphene.UUID(description="Added in 23.09.8.")
    session_owner = graphene.UUID(
        deprecation_reason="Deprecated since 23.09.8. use `session_owner_id`"
    )
    session_owner_email = graphene.String(description="Added in 23.09.8.")
    session_owner_id = graphene.UUID(description="Added in 23.09.8.")
    tag = graphene.String()
    startup_command = graphene.String()
    bootstrap_script = graphene.String()
    callback_url = graphene.String()
    environ = graphene.JSONString()
    name = graphene.String()
    resource_opts = graphene.JSONString()
    replicas = graphene.Int(description="Added in 24.12.0. Replaces `desired_session_count`.")
    desired_session_count = graphene.Int(
        deprecation_reason="Deprecated since 24.12.0. Use `replicas` instead."
    )
    cluster_mode = graphene.String()
    cluster_size = graphene.Int()
    open_to_public = graphene.Boolean()
    runtime_variant = graphene.Field(RuntimeVariantInfo, description="Added in 24.03.5.")

    created_at = GQLDateTime(required=True)
    destroyed_at = GQLDateTime()

    routings = graphene.List(Routing)
    retries = graphene.Int()
    status = graphene.String()

    lifecycle_stage = graphene.String()

    errors = graphene.List(graphene.NonNull(InferenceSessionError), required=True)

    live_stat = graphene.JSONString(description="Added in 24.12.0.")

    _queryfilter_fieldspec: Mapping[str, FieldSpecItem] = {
        "name": ("endpoints_name", None),
        "endpoint_id": ("endpoints_id", None),
        "project": ("endpoints_project", None),
        "resource_group": ("endpoints_resource_group", None),
        "created_at": ("endpoints_created_at", dtparse),
        "retries": ("endpoints_retries", None),
        "replicas": ("endpoints_replicas", None),
        "destroyed_at": ("endpoints_destroyed_at", dtparse),
        "model": ("endpoints_model", None),
        "domain": ("endpoints_domain", None),
        "url": ("endpoints_url", None),
        "lifecycle_stage": ("endpoints_lifecycle_stage", EndpointLifecycle),
        "open_to_public": ("endpoints_open_to_public", None),
        "created_user_id": ("users_id", None),
        "created_user_email": ("users_email", None),
    }

    _queryorder_colmap: Mapping[str, OrderSpecItem] = {
        "name": ("endpoints_name", None),
        "endpoint_id": ("endpoints_id", None),
        "project": ("endpoints_project", None),
        "resource_group": ("endpoints_resource_group", None),
        "retries": ("endpoints_retries", None),
        "replicas": ("endpoints_replicas", None),
        "created_at": ("endpoints_created_at", None),
        "destroyed_at": ("endpoints_destroyed_at", None),
        "model": ("endpoints_model", None),
        "domain": ("endpoints_domain", None),
        "url": ("endpoints_url", None),
        "lifecycle_stage": ("endpoints_lifecycle_stage", None),
        "open_to_public": ("endpoints_open_to_public", None),
        "created_user_id": ("users_id", None),
        "created_user_email": ("users_email", None),
    }

    @classmethod
    async def from_row(
        cls,
        ctx: GraphQueryContext,
        row: EndpointRow,
    ) -> Self:
        creator = cast(UserRow | None, row.created_user_row)
        return cls(
            endpoint_id=row.id,
            # image="", # deprecated, row.image_object.name,
            image_object=ImageNode.from_row(ctx, row.image_row),
            domain=row.domain,
            project=row.project,
            resource_group=row.resource_group,
            resource_slots=row.resource_slots.to_json(),
            url=row.url,
            model=row.model,
            model_definition_path=row.model_definition_path,
            model_mount_destiation=row.model_mount_destination,
            model_mount_destination=row.model_mount_destination,
            created_user=row.created_user,
            created_user_id=row.created_user,
            created_user_email=creator.email if creator is not None else None,
            session_owner=row.session_owner,
            session_owner_id=row.session_owner,
            session_owner_email=row.session_owner_row.email
            if row.session_owner_row is not None
            else None,
            tag=row.tag,
            startup_command=row.startup_command,
            bootstrap_script=row.bootstrap_script,
            callback_url=row.callback_url,
            environ=row.environ,
            name=row.name,
            resource_opts=row.resource_opts,
            replicas=row.replicas,
            desired_session_count=row.replicas,
            cluster_mode=row.cluster_mode,
            cluster_size=row.cluster_size,
            open_to_public=row.open_to_public,
            created_at=row.created_at,
            destroyed_at=row.destroyed_at,
            retries=row.retries,
            routings=[await Routing.from_row(ctx, r, endpoint=row) for r in row.routings],
            lifecycle_stage=row.lifecycle_stage.name,
            runtime_variant=RuntimeVariantInfo.from_enum(row.runtime_variant),
        )

    @classmethod
    def from_dto(cls, ctx: GraphQueryContext, dto: EndpointData | None) -> Self | None:
        if dto is None:
            return None
        return cls(
            endpoint_id=dto.id,
            image_object=ImageNode.from_row(ctx, ImageRow.from_optional_dataclass(dto.image)),
            domain=dto.domain,
            project=dto.project,
            resource_group=dto.resource_group,
            resource_slots=dto.resource_slots.to_json(),
            url=dto.url,
            model=dto.model,
            model_definition_path=dto.model_definition_path,
            model_mount_destiation=dto.model_mount_destination,
            model_mount_destination=dto.model_mount_destination,
            created_user=dto.created_user_id,
            created_user_id=dto.created_user_id,
            created_user_email=dto.created_user_email,
            session_owner=dto.session_owner_id,
            session_owner_id=dto.session_owner_id,
            session_owner_email=dto.session_owner_email,
            tag=dto.tag,
            startup_command=dto.startup_command,
            bootstrap_script=dto.bootstrap_script,
            callback_url=dto.callback_url,
            environ=dto.environ,
            name=dto.name,
            resource_opts=dto.resource_opts,
            replicas=dto.replicas,
            desired_session_count=dto.replicas,
            cluster_mode=dto.cluster_mode,
            cluster_size=dto.cluster_size,
            open_to_public=dto.open_to_public,
            created_at=dto.created_at,
            destroyed_at=dto.destroyed_at,
            retries=dto.retries,
            routings=[Routing.from_dto(r) for r in dto.routings] if dto.routings else None,
            lifecycle_stage=dto.lifecycle_stage,
            runtime_variant=RuntimeVariantInfo.from_enum(dto.runtime_variant),
        )

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        *,
        project: UUID | None = None,
        domain_name: str | None = None,
        user_uuid: UUID | None = None,
        filter: str | None = None,
    ) -> int:
        query = sa.select(sa.func.count()).select_from(
            sa.join(
                EndpointRow,
                UserRow,
                EndpointRow.session_owner == UserRow.uuid,
                isouter=True,
            )
        )
        if project is not None:
            query = query.where(EndpointRow.project == project)
        if domain_name is not None:
            query = query.where(EndpointRow.domain == domain_name)
        if user_uuid is not None:
            query = query.where(EndpointRow.session_owner == user_uuid)
        if filter is not None:
            filter_parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = filter_parser.append_filter(query, filter)

        async with ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar() or 0

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: str | None = None,
        user_uuid: UUID | None = None,
        project: UUID | None = None,
        filter: str | None = None,
        order: str | None = None,
    ) -> Sequence[Self]:
        query = (
            sa.select(EndpointRow)
            .select_from(
                sa.join(
                    EndpointRow,
                    UserRow,
                    EndpointRow.session_owner == UserRow.uuid,
                    isouter=True,
                )
            )
            .limit(limit)
            .offset(offset)
            .options(selectinload(EndpointRow.image_row).selectinload(ImageRow.aliases))
            .options(selectinload(EndpointRow.routings))
            .options(selectinload(EndpointRow.session_owner_row))
            .options(joinedload(EndpointRow.created_user_row))
        )
        if project is not None:
            query = query.where(EndpointRow.project == project)
        if domain_name is not None:
            query = query.where(EndpointRow.domain == domain_name)
        if user_uuid is not None:
            query = query.where(EndpointRow.session_owner == user_uuid)

        if filter is not None:
            filter_parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = filter_parser.append_filter(query, filter)
        if order is not None:
            order_parser = QueryOrderParser(cls._queryorder_colmap)
            query = order_parser.append_ordering(query, order)
        else:
            query = query.order_by(sa.desc(EndpointRow.created_at))

        async with ctx.db.begin_readonly_session() as db_session:
            result = await db_session.execute(query)
            return [await cls.from_row(ctx, row) for row in result.scalars().all()]

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        *,
        domain_name: str | None = None,
        user_uuid: UUID | None = None,
        project: UUID | None = None,
    ) -> Sequence[Self]:
        async with ctx.db.begin_readonly_session() as session:
            rows = await EndpointRow.list(
                session,
                project=project,
                domain=domain_name,
                user_uuid=user_uuid,
                load_image=True,
                load_created_user=True,
                load_session_owner=True,
            )
            return [await cls.from_row(ctx, row) for row in rows]

    @classmethod
    async def load_item(
        cls,
        ctx: GraphQueryContext,
        *,
        endpoint_id: UUID,
        domain_name: str | None = None,
        user_uuid: UUID | None = None,
        project: UUID | None = None,
    ) -> Self:
        """
        :raises: ai.backend.manager.api.exceptions.EndpointNotFound
        """
        try:
            async with ctx.db.begin_readonly_session() as session:
                row = await EndpointRow.get(
                    session,
                    endpoint_id=endpoint_id,
                    domain=domain_name,
                    user_uuid=user_uuid,
                    project=project,
                    load_image=True,
                    load_routes=True,
                    load_created_user=True,
                    load_session_owner=True,
                )
                return await cls.from_row(ctx, row)
        except NoResultFound as e:
            raise EndpointNotFound from e

    async def resolve_status(self, info: graphene.ResolveInfo) -> str:
        match self.lifecycle_stage:
            case EndpointLifecycle.DESTROYED.name:
                return EndpointStatus.DESTROYED
            case EndpointLifecycle.DESTROYING.name:
                return EndpointStatus.DESTROYING
            case _:
                active_route_status_names = {s.name for s in RouteStatus.active_route_statuses()}
                active_routings = [
                    r for r in self.routings if r.status in active_route_status_names
                ]
                healthy_count = sum(
                    1 for r in active_routings if r.status == RouteStatus.HEALTHY.name
                )
                if healthy_count == 0:
                    return EndpointStatus.UNHEALTHY
                if healthy_count == len(active_routings):
                    return EndpointStatus.HEALTHY
                return EndpointStatus.DEGRADED

    async def resolve_model_vfolder(self, info: graphene.ResolveInfo) -> VirtualFolderNode:
        if not self.model:
            raise ObjectNotFound(object_name="VFolder")

        ctx: GraphQueryContext = info.context

        async with ctx.db.begin_readonly_session() as sess:
            vfolder_row = await VFolderRow.get(sess, self.model, load_user=True, load_group=True)
            return VirtualFolderNode.from_row(info, vfolder_row)

    async def resolve_extra_mounts(self, info: graphene.ResolveInfo) -> Sequence[VirtualFolderNode]:
        if not self.endpoint_id:
            raise ObjectNotFound(object_name="Endpoint")

        ctx: GraphQueryContext = info.context

        async with ctx.db.begin_readonly_session() as sess:
            endpoint_row = await EndpointRow.get(sess, self.endpoint_id)
            extra_mount_folder_ids = [m.vfid.folder_id for m in endpoint_row.extra_mounts]
            query = (
                sa.select(VFolderRow)
                .where(VFolderRow.id.in_(extra_mount_folder_ids))
                .options(selectinload(VFolderRow.user_row))
                .options(selectinload(VFolderRow.group_row))
            )
            return [VirtualFolderNode.from_row(info, r) for r in (await sess.scalars(query))]

    async def resolve_errors(self, info: graphene.ResolveInfo) -> Any:
        error_routes = [r for r in self.routings if r.status == RouteStatus.FAILED_TO_START.name]
        errors = []
        for route in error_routes:
            if not route.error_data:
                continue
            match route.error_data["type"]:
                case "session_cancelled":
                    session_id = route.error_data["session_id"]
                case _:
                    session_id = None
            errors.append(
                InferenceSessionError(
                    session_id=session_id,
                    errors=[
                        InferenceSessionError.InferenceSessionErrorInfo(
                            src=e["src"], name=e["name"], repr=e["repr"]
                        )
                        for e in route.error_data["errors"]
                    ],
                )
            )

        return errors

    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Mapping[str, Any] | None:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(
            graph_ctx, "EndpointStatistics.by_endpoint"
        )
        return cast(Mapping[str, Any] | None, await loader.load(self.endpoint_id))


class EndpointList(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(Endpoint, required=True)


class ExtraMountInput(graphene.InputObjectType):  # type: ignore[misc]
    """Added in 24.03.4."""

    vfolder_id = graphene.String()
    mount_destination = graphene.String()
    type = graphene.String(
        description=f"Added in 24.03.4. Set bind type of this mount. Shoud be one of ({','.join([type_.value for type_ in MountTypes])}). Default is 'bind'."
    )
    permission = graphene.String(
        description=f"Added in 24.03.4. Set permission of this mount. Should be one of ({','.join([perm.value for perm in MountPermission])}). Default is null"
    )

    def to_action_field(self, info: graphene.ResolveInfo) -> ExtraMount:
        _, raw_vfolder_id = AsyncNode.resolve_global_id(info, self.vfolder_id)
        if not raw_vfolder_id:
            raw_vfolder_id = self.vfolder_id
        return ExtraMount(
            vfolder_id=OptionalState.from_graphql(UUID(raw_vfolder_id)),
            mount_destination=OptionalState.from_graphql(self.mount_destination),
            type=OptionalState.from_graphql(self.type),
            permission=OptionalState.from_graphql(self.permission),
        )


class ModifyEndpointInput(graphene.InputObjectType):  # type: ignore[misc]
    resource_slots = graphene.JSONString()
    resource_opts = graphene.JSONString()
    cluster_mode = graphene.String()
    cluster_size = graphene.Int()
    replicas = graphene.Int(description="Added in 24.12.0. Replaces `desired_session_count`.")
    desired_session_count = graphene.Int(
        deprecation_reason="Deprecated since 24.12.0. Use `replicas` instead."
    )
    image = ImageRefType()
    name = graphene.String()
    resource_group = graphene.String()
    model_definition_path = graphene.String(
        description="Added in 24.03.4. Must be set to `/models` when choosing `runtime_variant` other than `CUSTOM` or `CMD`."
    )
    open_to_public = graphene.Boolean()
    extra_mounts = graphene.List(
        ExtraMountInput,
        description="Added in 24.03.4. MODEL type VFolders are not allowed to be attached to model service session with this option.",
    )
    environ = graphene.JSONString(description="Added in 24.03.5.")
    runtime_variant = graphene.String(description="Added in 24.03.5.")

    def to_action(self, endpoint_id: uuid.UUID, info: graphene.ResolveInfo) -> ModifyEndpointAction:
        def create_image_ref_from_input(graphene_image_input: ImageRefType) -> ImageRef:
            registry: OptionalState[str] = OptionalState.nop()
            if (
                graphene_image_input.registry is not Undefined
                and graphene_image_input.registry is not None
            ):
                registry = OptionalState.update(graphene_image_input.registry)

            architecture: OptionalState[str] = OptionalState.nop()
            if (
                graphene_image_input.architecture is not Undefined
                and graphene_image_input.architecture is not None
            ):
                architecture = OptionalState.update(graphene_image_input.architecture)

            return ImageRef(graphene_image_input.name, registry, architecture)

        def convert_runtime_variant(
            value: str | None | UndefinedType,
        ) -> RuntimeVariant | UndefinedType:
            if isinstance(value, UndefinedType):
                return value
            if value is None:
                raise InvalidAPIParameters("Runtime variant cannot be None")

            try:
                return RuntimeVariant(value)
            except KeyError as e:
                raise InvalidAPIParameters(f"Unsupported runtime {self.runtime_variant}") from e

        if self.desired_session_count is not Undefined and self.replicas is not Undefined:
            raise InvalidAPIParameters(
                "Cannot set both desired_session_count and replicas. Use replicas for future use."
            )

        def convert_extra_mounts(
            extra_mounts_gql: list[ExtraMountInput] | UndefinedType,
        ) -> list[ExtraMount] | UndefinedType:
            if isinstance(extra_mounts_gql, UndefinedType):
                return extra_mounts_gql
            if extra_mounts_gql is None:
                raise InvalidAPIParameters("Extra mounts cannot be None")

            return [extra_mount.to_action_field(info) for extra_mount in extra_mounts_gql]

        spec = EndpointUpdaterSpec(
            resource_slots=OptionalState.from_graphql(
                self.resource_slots
                if (self.resource_slots is Undefined or self.resource_slots is None)
                else ResourceSlot.from_user_input(self.resource_slots, None)
            ),
            resource_opts=TriState.from_graphql(self.resource_opts),
            cluster_mode=OptionalState.from_graphql(
                self.cluster_mode
                if (self.cluster_mode is Undefined or self.cluster_mode is None)
                else ClusterMode(self.cluster_mode)
            ),
            cluster_size=OptionalState.from_graphql(self.cluster_size),
            replicas=OptionalState.from_graphql(self.replicas),
            desired_session_count=OptionalState.from_graphql(self.desired_session_count),
            image=TriState.from_graphql(
                create_image_ref_from_input(self.image) if self.image is not Undefined else None,
            ),
            name=OptionalState.from_graphql(self.name),
            resource_group=OptionalState.from_graphql(self.resource_group),
            model_definition_path=TriState.from_graphql(self.model_definition_path),
            open_to_public=OptionalState.from_graphql(
                self.open_to_public,
            ),
            extra_mounts=OptionalState.from_graphql(
                self.extra_mounts
                if (self.extra_mounts is Undefined or self.extra_mounts is None)
                else convert_extra_mounts(self.extra_mounts),
            ),
            environ=TriState.from_graphql(
                self.environ,
            ),
            runtime_variant=OptionalState.from_graphql(
                convert_runtime_variant(self.runtime_variant),
            ),
        )
        return ModifyEndpointAction(
            endpoint_id=endpoint_id,
            updater=Updater(spec=spec, pk_value=endpoint_id),
        )


class ModifyEndpoint(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.USER, UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        endpoint_id = graphene.UUID(required=True)
        props = ModifyEndpointInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    endpoint = graphene.Field(lambda: Endpoint, required=False, description="Added in 23.09.8.")

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        endpoint_id: UUID,
        props: ModifyEndpointInput,
    ) -> Self:
        graph_ctx: GraphQueryContext = info.context

        action = props.to_action(
            endpoint_id=endpoint_id,
            info=info,
        )

        result = await graph_ctx.processors.model_serving.modify_endpoint.wait_for_complete(action)

        return cls(
            ok=result.success,
            msg="success" if result.success else "failed",
            endpoint=Endpoint.from_dto(graph_ctx, result.data),
        )


@graphene_federation.key("token")
class EndpointToken(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (Item,)

    token = graphene.String(required=True)
    endpoint_id = graphene.UUID(required=True)
    domain = graphene.String(required=True)
    project = graphene.String(required=True)
    session_owner = graphene.UUID(required=True)

    created_at = GQLDateTime(required=True)
    valid_until = GQLDateTime()

    @classmethod
    async def from_row(
        cls,
        ctx: GraphQueryContext,
        row: EndpointTokenRow,
    ) -> Self:
        return cls(
            token=row.token,
            endpoint_id=row.endpoint,
            domain=row.domain,
            project=row.project,
            session_owner=row.session_owner,
            created_at=row.created_at,
        )

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        *,
        endpoint_id: UUID | None = None,
        project: UUID | None = None,
        domain_name: str | None = None,
        user_uuid: UUID | None = None,
    ) -> int:
        query = sa.select(sa.func.count()).select_from(EndpointTokenRow)
        if endpoint_id is not None:
            query = query.where(EndpointTokenRow.endpoint == endpoint_id)
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain_name:
            query = query.filter(EndpointTokenRow.domain == domain_name)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        async with ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar() or 0

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        endpoint_id: UUID | None = None,
        filter: str | None = None,
        order: str | None = None,
        project: UUID | None = None,
        domain_name: str | None = None,
        user_uuid: UUID | None = None,
    ) -> Sequence[Self]:
        query = (
            sa.select(EndpointTokenRow)
            .limit(limit)
            .offset(offset)
            .order_by(sa.desc(EndpointTokenRow.created_at))
        )
        if endpoint_id is not None:
            query = query.where(EndpointTokenRow.endpoint == endpoint_id)
        if project:
            query = query.filter(EndpointTokenRow.project == project)
        if domain_name:
            query = query.filter(EndpointTokenRow.domain == domain_name)
        if user_uuid:
            query = query.filter(EndpointTokenRow.session_owner == user_uuid)
        """
        if filter is not None:
            parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = parser.append_filter(query, filter)
        if order is not None:
            parser = QueryOrderParser(cls._queryorder_colmap)
            query = parser.append_ordering(query, order)
        """
        async with ctx.db.begin_readonly_session() as session:
            result = await session.execute(query)
            return [await cls.from_row(ctx, row) for row in result.scalars().all()]

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        endpoint_id: UUID,
        *,
        project: UUID | None = None,
        domain_name: str | None = None,
        user_uuid: UUID | None = None,
    ) -> Sequence[Self]:
        async with ctx.db.begin_readonly_session() as session:
            rows = await EndpointTokenRow.list(
                session,
                endpoint_id,
                project=project,
                domain=domain_name,
                user_uuid=user_uuid,
            )
        return [await cls.from_row(ctx, row) for row in rows]

    @classmethod
    async def load_item(
        cls,
        ctx: GraphQueryContext,
        token: str,
        *,
        project: UUID | None = None,
        domain_name: str | None = None,
        user_uuid: UUID | None = None,
    ) -> Self:
        try:
            async with ctx.db.begin_readonly_session() as session:
                row = await EndpointTokenRow.get(
                    session, token, project=project, domain=domain_name, user_uuid=user_uuid
                )
        except NoResultFound as e:
            raise EndpointTokenNotFound from e
        return await cls.from_row(ctx, row)

    async def resolve_valid_until(
        self,
        info: graphene.ResolveInfo,
    ) -> datetime.datetime | None:
        try:
            decoded = jwt.decode(
                self.token,
                algorithms=["HS256"],
                options={"verify_signature": False, "verify_exp": False},
            )
        except jwt.DecodeError:
            return None
        if "exp" not in decoded:
            return None
        return datetime.datetime.fromtimestamp(decoded["exp"], tz=tzutc())


class EndpointTokenList(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(EndpointToken, required=True)



import logging
from collections.abc import AsyncIterator, Mapping
from contextlib import asynccontextmanager as actxmgr
from dataclasses import dataclass
from http import HTTPStatus
from typing import Any, Final, cast

import aiohttp
import yarl
from aiohttp import ClientTimeout

from ai.backend.common.exception import (
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
    InvalidErrorCode,
    PassthroughError,
)
from ai.backend.common.json import load_json
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.storage import (
    QuotaScopeNotFoundError,
    StorageProxyConnectionError,
    StorageProxyTimeoutError,
    UnexpectedStorageProxyResponseError,
    VFolderBadRequest,
    VFolderGone,
    VFolderNotFound,
    VFolderOperationFailed,
)

AUTH_TOKEN_HDR: Final = "X-BackendAI-Storage-Auth-Token"
DEFAULT_TIMEOUT: Final = ClientTimeout(total=300, sock_connect=30)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class StorageProxyClientArgs:
    endpoint: yarl.URL
    secret: str


class StorageProxyHTTPClient:
    _client_session: aiohttp.ClientSession
    _endpoint: yarl.URL
    _secret: str

    def __init__(self, client_session: aiohttp.ClientSession, args: StorageProxyClientArgs) -> None:
        self._client_session = client_session
        self._endpoint = args.endpoint
        self._secret = args.secret

    def _handle_vfolder_failure(self, status_code: HTTPStatus) -> None:
        match status_code:
            case HTTPStatus.BAD_REQUEST:
                raise VFolderBadRequest(
                    extra_msg="Bad request to storage proxy",
                )
            case HTTPStatus.NOT_FOUND:
                raise VFolderNotFound(
                    extra_msg="Requested resource not found",
                )
            case HTTPStatus.GONE:
                raise VFolderGone(
                    extra_msg=(
                        "The requested resource is gone. It may have been deleted or moved."
                    ),
                )
            case HTTPStatus.INTERNAL_SERVER_ERROR:
                raise VFolderOperationFailed(
                    extra_msg="Internal server error from storage proxy",
                )
            case _:
                raise VFolderOperationFailed(
                    extra_msg=f"Unexpected error {status_code} from storage proxy",
                )

    def _handle_quota_scope_failure(self, status_code: HTTPStatus) -> None:
        match status_code:
            case HTTPStatus.NOT_FOUND:
                raise QuotaScopeNotFoundError(
                    extra_msg="Requested quota scope not found",
                )
            case _:
                raise UnexpectedStorageProxyResponseError(
                    extra_msg=f"Unexpected error {status_code} from storage proxy",
                )

    async def _handle_exceptional_response(self, resp: aiohttp.ClientResponse) -> None:
        data = None
        try:
            data = await resp.json()
        except (aiohttp.ContentTypeError, ValueError) as e:
            resp_text = await resp.text()
            log.warning(
                "Failed to parse JSON from storage proxy error response: "
                "status={}, content_type={}, error={}, response_text={}",
                resp.status,
                resp.content_type,
                e,
                resp_text if resp_text else "",
            )
            raise PassthroughError(
                status_code=resp.status,
                error_code=ErrorCode(
                    domain=ErrorDomain.STORAGE_PROXY,
                    operation=ErrorOperation.REQUEST,
                    error_detail=ErrorDetail.CONTENT_TYPE_MISMATCH,
                ),
                error_message=f"Failed to parse error response from storage proxy. Original response: {resp_text if resp_text else ''}",
            ) from e
        try:
            err_code = ErrorCode.from_str(data.get("error_code", ""))
            err_domain = err_code.domain
        except InvalidErrorCode:
            err_domain = ErrorDomain.VFOLDER  # Default domain if parsing fails
        match err_domain:
            case ErrorDomain.VFOLDER:
                self._handle_vfolder_failure(HTTPStatus(resp.status))
            case ErrorDomain.QUOTA_SCOPE:
                self._handle_quota_scope_failure(HTTPStatus(resp.status))
            case _:
                raise UnexpectedStorageProxyResponseError(
                    extra_msg=f"Unexpected error {resp.status} from storage proxy",
                )

    @actxmgr
    async def request_stream_response(
        self,
        method: str,
        url: str,
        *,
        body: Mapping[str, Any] | None = None,
        params: Mapping[str, Any] | None = None,
        request_timeout: ClientTimeout,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        """
        Make an HTTP request using the session client.

        :param method: HTTP method (GET, POST, etc.)
        :param url: URL to send the request to
        :param body: JSON body data to send with the request
        :param request_timeout: Timeout configuration for the request
        :return: Response data as a dictionary, or None if no content
        """
        headers = {
            AUTH_TOKEN_HDR: self._secret,
        }
        try:
            async with self._client_session.request(
                method,
                self._endpoint / url,
                headers=headers,
                json=body,
                params=params,
                timeout=request_timeout,
            ) as client_resp:
                if client_resp.status // 100 == 2:
                    yield client_resp
                    return
                await self._handle_exceptional_response(client_resp)
        except TimeoutError as e:
            raise StorageProxyTimeoutError(
                extra_msg="Request to storage proxy timed out",
            ) from e
        except aiohttp.ClientConnectionError as e:
            raise StorageProxyConnectionError(
                extra_msg="Failed to connect to storage proxy",
            ) from e

    async def request(
        self,
        method: str,
        url: str,
        *,
        body: Mapping[str, Any] | None = None,
        params: Mapping[str, Any] | None = None,
        request_timeout: ClientTimeout,
    ) -> Mapping[str, Any] | None:
        """
        Make an HTTP request using the session client.

        :param method: HTTP method (GET, POST, etc.)
        :param url: URL to send the request to
        :param body: JSON body data to send with the request
        :param request_timeout: Timeout configuration for the request
        :return: Response data as a dictionary, or None if no content
        """
        async with self.request_stream_response(
            method, url, body=body, params=params, request_timeout=request_timeout
        ) as response_stream:
            if response_stream.status == HTTPStatus.NO_CONTENT:
                return None
            resp_bytes = await response_stream.read()
            if not resp_bytes:
                return None
            return cast(Mapping[str, Any] | None, load_json(resp_bytes))

    async def request_with_response(
        self,
        method: str,
        url: str,
        *,
        body: Mapping[str, Any] | None = None,
        params: Mapping[str, Any] | None = None,
        request_timeout: ClientTimeout,
    ) -> Mapping[str, Any]:
        """
        Make an HTTP request and return the response as a dictionary.

        :param method: HTTP method (GET, POST, etc.)
        :param url: URL to send the request to
        :param body: JSON body data to send with the request
        :param request_timeout: Timeout configuration for the request
        :return: Response object from the request
        """
        response = await self.request(
            method, url, body=body, params=params, request_timeout=request_timeout
        )
        if response is None:
            raise UnexpectedStorageProxyResponseError(
                "Unexpected response from storage proxy: None",
            )
        return response



from .create import CreateVFSStorageAction, CreateVFSStorageActionResult
from .delete import DeleteVFSStorageAction, DeleteVFSStorageActionResult
from .get import GetVFSStorageAction, GetVFSStorageActionResult
from .list import ListVFSStorageAction, ListVFSStorageActionResult
from .search import SearchVFSStoragesAction, SearchVFSStoragesActionResult
from .update import UpdateVFSStorageAction, UpdateVFSStorageActionResult

__all__ = [
    "CreateVFSStorageAction",
    "CreateVFSStorageActionResult",
    "DeleteVFSStorageAction",
    "DeleteVFSStorageActionResult",
    "GetVFSStorageAction",
    "GetVFSStorageActionResult",
    "ListVFSStorageAction",
    "ListVFSStorageActionResult",
    "SearchVFSStoragesAction",
    "SearchVFSStoragesActionResult",
    "UpdateVFSStorageAction",
    "UpdateVFSStorageActionResult",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.notification import NotificationChannelData

from .base import NotificationAction


@dataclass
class GetChannelAction(NotificationAction):
    """Action to get a notification channel by ID."""

    channel_id: UUID

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET

    @override
    def entity_id(self) -> str | None:
        return str(self.channel_id)


@dataclass
class GetChannelActionResult(BaseActionResult):
    """Result of getting a notification channel."""

    channel_data: NotificationChannelData

    @override
    def entity_id(self) -> str | None:
        return str(self.channel_data.id)



"""Export API request DTOs for path parameters and headers."""

from pydantic import Field

from ai.backend.common.api_handlers import BaseRequestModel


class ExportPathParam(BaseRequestModel):
    """Path parameter for export endpoints."""

    report_key: str = Field(
        description="The report key to export (e.g., 'sessions', 'users', 'projects')"
    )


class ExportFilenameHeader(BaseRequestModel):
    """Header parameter for optional export filename."""

    filename: str | None = Field(
        default=None,
        validation_alias="X-Export-Filename",
        description="Optional filename for the exported CSV file",
    )



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.user.types import BulkUserCreateResultData, UserCreateResultData
from ai.backend.manager.models.user import UserRow
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.services.user.actions.base import UserAction


@dataclass
class CreateUserAction(UserAction):
    creator: Creator[UserRow]
    group_ids: list[str] | None = None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class CreateUserActionResult(BaseActionResult):
    data: UserCreateResultData

    @override
    def entity_id(self) -> str | None:
        return str(self.data.user.id)


@dataclass
class UserCreateSpec:
    """Specification for creating a single user, including group assignments."""

    creator: Creator[UserRow]
    group_ids: list[str] | None = None


@dataclass
class BulkCreateUserAction(UserAction):
    """Action for bulk creating multiple users."""

    items: list[UserCreateSpec]

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class BulkCreateUserActionResult(BaseActionResult):
    """Result of bulk user creation."""

    data: BulkUserCreateResultData

    @override
    def entity_id(self) -> str | None:
        return None



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.group.repository import GroupRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class GroupRepositories:
    repository: GroupRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = GroupRepository(
            args.db,
            args.config_provider,
            args.valkey_stat_client,
            args.storage_manager,
        )
        return cls(repository=repository)



"""
Adapter for ObjectPermission RBAC operations.
Converts between API DTOs and service layer actions.
"""

from __future__ import annotations

import uuid

from ai.backend.common.dto.manager.rbac import (
    CreateObjectPermissionRequest,
    ObjectPermissionDTO,
)
from ai.backend.manager.data.permission.object_permission import ObjectPermissionData
from ai.backend.manager.models.rbac_models.permission.object_permission import ObjectPermissionRow
from ai.backend.manager.repositories.base import Creator, Purger
from ai.backend.manager.repositories.permission_controller.creators import (
    ObjectPermissionCreatorSpec,
)
from ai.backend.manager.services.permission_contoller.actions.object_permission import (
    CreateObjectPermissionAction,
    DeleteObjectPermissionAction,
)

__all__ = ("ObjectPermissionAdapter",)


class ObjectPermissionAdapter:
    """Adapter for converting object permission requests to actions and data to DTOs."""

    def to_object_permission_dto(self, data: ObjectPermissionData) -> ObjectPermissionDTO:
        """Convert ObjectPermissionData to ObjectPermissionDTO."""
        return ObjectPermissionDTO(
            id=data.id,
            role_id=data.role_id,
            entity_type=data.object_id.entity_type,
            entity_id=data.object_id.entity_id,
            operation=data.operation,
        )

    def to_create_object_permission_action(
        self,
        request: CreateObjectPermissionRequest,
    ) -> CreateObjectPermissionAction:
        """Convert CreateObjectPermissionRequest to CreateObjectPermissionAction."""
        creator = Creator(
            spec=ObjectPermissionCreatorSpec(
                role_id=request.role_id,
                entity_type=request.entity_type,
                entity_id=request.entity_id,
                operation=request.operation,
                status=request.status,
            )
        )
        return CreateObjectPermissionAction(creator=creator)

    def to_delete_object_permission_action(
        self,
        object_permission_id: uuid.UUID,
    ) -> DeleteObjectPermissionAction:
        """Convert object_permission_id to DeleteObjectPermissionAction."""
        purger = Purger(row_class=ObjectPermissionRow, pk_value=object_permission_id)
        return DeleteObjectPermissionAction(purger=purger)



from ai.backend.common.types import SessionId, SessionResult, SessionTypes
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.models.types import QueryCondition, QueryOption

from .row import (
    AGENT_RESOURCE_OCCUPYING_SESSION_STATUSES,
    ALLOWED_IMAGE_ROLES_FOR_SESSION_TYPE,
    DEAD_SESSION_STATUSES,
    DEFAULT_SESSION_ORDERING,
    PRIVATE_SESSION_TYPES,
    SESSION_KERNEL_STATUS_MAPPING,
    SESSION_STATUS_TRANSITION_MAP,
    USER_RESOURCE_OCCUPYING_SESSION_STATUSES,
    ConcurrencyUsed,
    KernelLoadingStrategy,
    SessionDependencyRow,
    SessionLifecycleManager,
    SessionRow,
    by_domain_name,
    by_raw_filter,
    by_resource_group_name,
    by_status,
    by_user_id,
    check_all_dependencies,
    determine_session_status_by_kernels,
    handle_session_exception,
)
from .row import (
    _build_session_fetch_query as _build_session_fetch_query,
)
from .row import (
    get_permission_ctx as get_permission_ctx,
)

__all__ = (
    "AGENT_RESOURCE_OCCUPYING_SESSION_STATUSES",
    "ALLOWED_IMAGE_ROLES_FOR_SESSION_TYPE",
    "DEAD_SESSION_STATUSES",
    "DEFAULT_SESSION_ORDERING",
    "PRIVATE_SESSION_TYPES",
    "SESSION_KERNEL_STATUS_MAPPING",
    "SESSION_STATUS_TRANSITION_MAP",
    "USER_RESOURCE_OCCUPYING_SESSION_STATUSES",
    "ConcurrencyUsed",
    "KernelLoadingStrategy",
    "QueryCondition",
    "QueryOption",
    "SessionDependencyRow",
    "SessionId",
    "SessionLifecycleManager",
    "SessionResult",
    "SessionRow",
    "SessionStatus",
    "SessionTypes",
    "by_domain_name",
    "by_raw_filter",
    "by_resource_group_name",
    "by_status",
    "by_user_id",
    "check_all_dependencies",
    "determine_session_status_by_kernels",
    "handle_session_exception",
)



"""Pagination strategies for repository queries."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa

from .types import QueryCondition, QueryOrder, TRow

if TYPE_CHECKING:
    from sqlalchemy.engine import Row


class QueryPagination(ABC):
    """
    Base class for pagination strategies.

    Subclasses must implement the apply() method to transform a SQLAlchemy
    select statement with appropriate pagination logic.
    """

    @property
    @abstractmethod
    def uses_window_function(self) -> bool:
        """Whether this pagination uses window function for total_count.

        Returns:
            True if window function should be added to main query (Offset),
            False if separate count query should be used (Cursor).
        """
        raise NotImplementedError

    @abstractmethod
    def apply(self, query: sa.sql.Select[Any]) -> sa.sql.Select[Any]:
        """Apply pagination to a SQLAlchemy select statement."""

        raise NotImplementedError

    @abstractmethod
    def compute_page_info(self, rows: list[Row[Any]], total_count: int) -> PageInfoResult[Row[Any]]:
        """Compute pagination info and slice rows if needed.

        Args:
            rows: The rows returned from query (may include extra row for page detection)
            total_count: Total count of items matching the query

        Returns:
            _PageInfoResult containing sliced rows and pagination flags
        """

        raise NotImplementedError


@dataclass
class PageInfoResult[TRow]:
    """Result of compute_page_info containing sliced rows and pagination flags."""

    rows: list[TRow]
    has_next_page: bool
    has_previous_page: bool


@dataclass
class NoPagination(QueryPagination):
    """
    No pagination - returns all matching rows.

    Use this when you need to fetch all results without any limit.
    Useful for internal operations like scheduler batch processing.
    """

    @property
    def uses_window_function(self) -> bool:
        return False

    def apply(self, query: sa.sql.Select[Any]) -> sa.sql.Select[Any]:
        """No pagination applied - returns query unchanged."""
        return query

    def compute_page_info(
        self, rows: list[Row[Any]], _total_count: int
    ) -> PageInfoResult[Row[Any]]:
        """No pagination - no next/previous pages."""
        return PageInfoResult(
            rows=rows,
            has_next_page=False,
            has_previous_page=False,
        )


@dataclass
class OffsetPagination(QueryPagination):
    """
    Offset-based pagination using limit and offset.

    This is the traditional SQL pagination approach where you specify:
    - limit: Maximum number of items to return
    - offset: Number of items to skip from the beginning
    """

    limit: int
    """Maximum number of items to return (must be positive)."""

    offset: int = 0
    """Number of items to skip from the beginning (must be non-negative)."""

    @property
    def uses_window_function(self) -> bool:
        return True

    def apply(self, query: sa.sql.Select[Any]) -> sa.sql.Select[Any]:
        """Apply offset-based pagination to query."""

        query = query.limit(self.limit)
        if self.offset > 0:
            query = query.offset(self.offset)
        return query

    def compute_page_info(self, rows: list[Row[Any]], total_count: int) -> PageInfoResult[Row[Any]]:
        """Compute pagination info for offset-based pagination."""

        has_previous_page = self.offset > 0
        has_next_page = (self.offset + len(rows)) < total_count
        return PageInfoResult(
            rows=rows,
            has_next_page=has_next_page,
            has_previous_page=has_previous_page,
        )


@dataclass
class CursorForwardPagination(QueryPagination):
    """
    Cursor-based forward pagination using first and after.

    This follows the GraphQL Relay Cursor Connections specification for forward pagination.
    Use this to paginate forward through a result set:
    - first: Number of items to return from the cursor position
    - cursor_condition: Optional QueryCondition for WHERE clause (e.g., id > cursor_id).
      If None, starts from the beginning.
    - cursor_order: QueryOrder for cursor-based ordering (e.g., created_at ASC)
    """

    first: int
    """Number of items to return (must be positive)."""

    cursor_order: QueryOrder
    """Ordering for cursor-based pagination (e.g., ORDER BY created_at ASC)."""

    cursor_condition: QueryCondition | None = None
    """Optional QueryCondition for cursor position. If None, starts from the beginning."""

    @property
    def uses_window_function(self) -> bool:
        return False

    def apply(self, query: sa.sql.Select[Any]) -> sa.sql.Select[Any]:
        """
        Apply cursor-based forward pagination to query.

        Applies cursor condition (if present), cursor order, and LIMIT + 1 for has_next_page detection.
        """
        if self.cursor_condition is not None:
            query = query.where(self.cursor_condition())
        query = query.order_by(self.cursor_order)
        return query.limit(self.first + 1)

    def compute_page_info(
        self, rows: list[Row[Any]], _total_count: int
    ) -> PageInfoResult[Row[Any]]:
        """Compute pagination info for cursor-based forward pagination."""

        # has_previous_page is True only if cursor was provided (meaning we're past the first page)
        has_previous_page = self.cursor_condition is not None
        has_next_page = len(rows) > self.first
        if has_next_page:
            rows = rows[: self.first]
        return PageInfoResult(
            rows=rows,
            has_next_page=has_next_page,
            has_previous_page=has_previous_page,
        )


@dataclass
class CursorBackwardPagination(QueryPagination):
    """
    Cursor-based backward pagination using last and before.

    This follows the GraphQL Relay Cursor Connections specification for backward pagination.
    Use this to paginate backward through a result set:
    - last: Number of items to return before the cursor position
    - cursor_condition: Optional QueryCondition for WHERE clause (e.g., id < cursor_id).
      If None, starts from the end.
    - cursor_order: QueryOrder for cursor-based ordering (e.g., created_at DESC for reverse fetch)
    """

    last: int
    """Number of items to return (must be positive)."""

    cursor_order: QueryOrder
    """Ordering for cursor-based pagination (e.g., ORDER BY created_at DESC)."""

    cursor_condition: QueryCondition | None = None
    """Optional QueryCondition for cursor position. If None, starts from the end."""

    @property
    def uses_window_function(self) -> bool:
        return False

    def apply(self, query: sa.sql.Select[Any]) -> sa.sql.Select[Any]:
        """
        Apply cursor-based backward pagination to query.

        Applies cursor condition (if present), cursor order, and LIMIT + 1 for has_previous_page detection.
        """
        if self.cursor_condition is not None:
            query = query.where(self.cursor_condition())
        query = query.order_by(self.cursor_order)
        return query.limit(self.last + 1)

    def compute_page_info(
        self, rows: list[Row[Any]], _total_count: int
    ) -> PageInfoResult[Row[Any]]:
        """Compute pagination info for cursor-based backward pagination."""

        # has_next_page is True only if cursor was provided (meaning there are items after this page)
        has_next_page = self.cursor_condition is not None
        has_previous_page = len(rows) > self.last
        if has_previous_page:
            rows = rows[: self.last]
        return PageInfoResult(
            rows=rows,
            has_next_page=has_next_page,
            has_previous_page=has_previous_page,
        )



from __future__ import annotations

import logging
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, cast

import aiohttp
import graphene

from ai.backend.common.types import HardwareMetadata
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.storage_proxy.session_manager import (
    AUTH_TOKEN_HDR,
    VolumeInfo,
)

from .base import Item, PaginatedList

if TYPE_CHECKING:
    from .schema import GraphQueryContext


__all__ = (
    "StorageVolume",
    "StorageVolumeList",
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class StorageVolume(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (Item,)

    # id: {proxy_name}:{name}
    backend = graphene.String()
    fsprefix = graphene.String()
    path = graphene.String()
    capabilities = graphene.List(graphene.String)
    hardware_metadata = graphene.JSONString()
    performance_metric = graphene.JSONString()
    usage = graphene.JSONString()
    proxy = graphene.String(
        description="Added in 24.03.0. Name of the proxy which this volume belongs to."
    )
    name = graphene.String(description="Added in 24.03.0. Name of the storage.")

    async def resolve_hardware_metadata(self, info: graphene.ResolveInfo) -> HardwareMetadata:
        ctx: GraphQueryContext = info.context
        return await ctx.registry.gather_storage_hwinfo(self.id)

    async def resolve_performance_metric(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        ctx: GraphQueryContext = info.context
        try:
            proxy_name, volume_name = ctx.storage_manager.get_proxy_and_volume(self.id)
            manager_client = ctx.storage_manager.get_manager_facing_client(proxy_name)
            storage_reply = await manager_client.get_volume_performance_metric(volume_name)
            return cast(Mapping[str, Any], storage_reply["metric"])
        except aiohttp.ClientResponseError:
            return {}

    async def resolve_usage(self, info: graphene.ResolveInfo) -> Mapping[str, Any]:
        ctx: GraphQueryContext = info.context
        proxy_name, volume_name = ctx.storage_manager.get_proxy_and_volume(self.id)
        try:
            manager_client = ctx.storage_manager.get_manager_facing_client(proxy_name)
            return await manager_client.get_fs_usage(volume_name)
        except aiohttp.ClientResponseError:
            return {}

    @classmethod
    def from_info(cls, proxy_name: str, volume_info: VolumeInfo) -> StorageVolume:
        return cls(
            id=f"{proxy_name}:{volume_info['name']}",
            backend=volume_info["backend"],
            path=volume_info["path"],
            fsprefix=volume_info["fsprefix"],
            capabilities=volume_info["capabilities"],
            name=volume_info["name"],
            proxy=proxy_name,
        )

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        filter: str | None = None,
    ) -> int:
        volumes = [*await ctx.storage_manager.get_all_volumes()]
        return len(volumes)

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        filter: str | None = None,
        order: str | None = None,
    ) -> Sequence[StorageVolume]:
        # For consistency we add filter/order params here, but it's actually noop.
        if filter is not None or order is not None:
            log.warning(
                "Paginated list of storage volumes igonores custom filtering and/or ordering"
            )
        volumes = [*await ctx.storage_manager.get_all_volumes()]
        return [
            cls.from_info(proxy_name, volume_info)
            for proxy_name, volume_info in volumes[offset : offset + limit]
        ]

    @classmethod
    async def load_by_id(
        cls,
        ctx: GraphQueryContext,
        id: str,
    ) -> StorageVolume:
        proxy_name, volume_name = ctx.storage_manager.get_proxy_and_volume(id)
        try:
            proxy_info = ctx.storage_manager._proxies[proxy_name]
        except KeyError as e:
            raise ValueError(f"no such storage proxy: {proxy_name!r}") from e
        async with proxy_info.session.request(
            "GET",
            proxy_info.manager_api_url / "volumes",
            raise_for_status=True,
            headers={AUTH_TOKEN_HDR: proxy_info.secret},
        ) as resp:
            reply = await resp.json()
            for volume_data in reply["volumes"]:
                if volume_data["name"] == volume_name:
                    return cls.from_info(proxy_name, volume_data)
            else:
                raise ValueError(
                    f"no such volume in the storage proxy {proxy_name!r}: {volume_name!r}",
                )


class StorageVolumeList(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(StorageVolume, required=True)



"""Handler for checking route health status."""

import logging
from collections.abc import Sequence

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.deployment.types import RouteStatus, RouteStatusTransitions
from ai.backend.manager.defs import LockID
from ai.backend.manager.repositories.deployment.types import RouteData
from ai.backend.manager.sokovan.deployment.route.executor import RouteExecutor
from ai.backend.manager.sokovan.deployment.route.types import RouteExecutionResult

from .base import RouteHandler

log = BraceStyleAdapter(logging.getLogger(__name__))


class HealthCheckRouteHandler(RouteHandler):
    """Handler for checking route health status (readiness and liveness)."""

    def __init__(
        self,
        route_executor: RouteExecutor,
        event_producer: EventProducer,
    ) -> None:
        self._route_executor = route_executor
        self._event_producer = event_producer

    @classmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        return "health-check-routes"

    @property
    def lock_id(self) -> LockID | None:
        """Lock for health check routes."""
        return LockID.LOCKID_DEPLOYMENT_HEALTH_CHECK_ROUTES

    @classmethod
    def target_statuses(cls) -> list[RouteStatus]:
        """Get the target route statuses for this handler."""
        return [RouteStatus.HEALTHY, RouteStatus.UNHEALTHY, RouteStatus.DEGRADED]

    @classmethod
    def next_status(cls) -> RouteStatus | None:
        """Routes that pass health check become HEALTHY."""
        return RouteStatus.HEALTHY

    @classmethod
    def failure_status(cls) -> RouteStatus | None:
        """Routes that fail health check become UNHEALTHY."""
        return RouteStatus.UNHEALTHY

    @classmethod
    def stale_status(cls) -> RouteStatus | None:
        """Routes with stale health data become DEGRADED."""
        return RouteStatus.DEGRADED

    @classmethod
    def status_transitions(cls) -> RouteStatusTransitions:
        """Define state transitions for health check route handler (BEP-1030).

        - success: Route → HEALTHY
        - failure: Route → UNHEALTHY
        - stale: Route → DEGRADED (stale health data)
        """
        return RouteStatusTransitions(
            success=RouteStatus.HEALTHY,
            failure=RouteStatus.UNHEALTHY,
            stale=RouteStatus.DEGRADED,
        )

    async def execute(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Execute health check for routes."""
        log.debug("Checking health for {} routes", len(routes))

        # Execute route health check logic via executor
        return await self._route_executor.check_route_health(routes)

    async def post_process(self, result: RouteExecutionResult) -> None:
        """Handle post-processing after health check."""
        log.debug(
            "Health check: {} healthy, {} unhealthy, {} degraded",
            len(result.successes),
            len(result.errors),
            len(result.stale),
        )



from dataclasses import dataclass
from typing import Any, override

from ai.backend.common.types import AccessKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.services.session.actions.app_service_base import SessionAppServiceAction


@dataclass
class StartServiceAction(SessionAppServiceAction):
    session_name: str
    access_key: AccessKey
    service: str
    login_session_token: Any
    port: int | None
    arguments: str | None  # json_string
    envs: str | None  # json_string

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class StartServiceActionResult(BaseActionResult):
    # TODO: Add proper type
    result: Any
    session_data: SessionData
    token: str
    wsproxy_addr: str

    @override
    def entity_id(self) -> str | None:
        return str(self.session_data.id)



from abc import abstractmethod
from typing import TypeVar, override

from ai.backend.common.data.permission.types import ScopeType

from .base import BaseAction, BaseActionResult


class BaseScopeAction(BaseAction):
    @override
    def entity_id(self) -> str | None:
        return None

    @abstractmethod
    def scope_type(self) -> ScopeType:
        raise NotImplementedError

    @abstractmethod
    def scope_id(self) -> str:
        raise NotImplementedError


class BaseScopeActionResult(BaseActionResult):
    @override
    def entity_id(self) -> str | None:
        return None

    @abstractmethod
    def scope_type(self) -> ScopeType:
        raise NotImplementedError

    @abstractmethod
    def scope_id(self) -> str:
        raise NotImplementedError


TScopeAction = TypeVar("TScopeAction", bound=BaseScopeAction)
TScopeActionResult = TypeVar("TScopeActionResult", bound=BaseScopeActionResult)



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.domain.types import DomainData, UserInfo
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class ModifyDomainNodeAction(DomainAction):
    user_info: UserInfo
    updater: Updater[DomainRow]
    sgroups_to_add: set[str] | None = None
    sgroups_to_remove: set[str] | None = None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class ModifyDomainNodeActionResult(BaseActionResult):
    domain_data: DomainData

    @override
    def entity_id(self) -> str:
        return self.domain_data.name



"""Data types for notification system."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING

from ai.backend.common.data.notification import NotifiableMessage, NotificationRuleType

if TYPE_CHECKING:
    from ai.backend.manager.data.notification import NotificationChannelData

__all__ = (
    "NotificationMessage",
    "ProcessRuleParams",
    "SendResult",
)


@dataclass(frozen=True)
class NotificationMessage:
    """Notification message to be sent through a channel."""

    message: str


@dataclass(frozen=True)
class ProcessRuleParams:
    """Parameters for processing a notification rule."""

    message_template: str
    rule_type: NotificationRuleType
    channel: NotificationChannelData
    timestamp: datetime
    notification_data: NotifiableMessage


@dataclass(frozen=True)
class SendResult:
    """Result of sending a notification."""

    message: str



"""create vfolder_invitations table

Revision ID: 26d0c387e764
Revises: f8a71c3bffa2
Create Date: 2018-07-04 14:57:46.517587

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "26d0c387e764"
down_revision = "f8a71c3bffa2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "vfolder_invitations",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("permission", sa.String(length=2), nullable=True),
        sa.Column("invitee", sa.String(length=256), nullable=False),
        sa.Column("vfolder", GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["vfolder"],
            ["vfolders.id"],
            name=op.f("fk_vfolder_invitations_vfolder_vfolders"),
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_vfolder_invitations")),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("vfolder_invitations")
    # ### end Alembic commands ###



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.image.repository import ImageRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class ImageRepositories:
    repository: ImageRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = ImageRepository(args.db, args.valkey_image_client, args.config_provider)

        return cls(
            repository=repository,
        )



from __future__ import annotations

import uuid
from dataclasses import dataclass
from pathlib import Path


@dataclass
class VFSStorageListResult:
    """Search result with total count for VFS storages."""

    items: list[VFSStorageData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass
class VFSStorageData:
    id: uuid.UUID
    name: str
    host: str
    base_path: Path



from __future__ import annotations

import logging
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any
from uuid import UUID

import graphene
import sqlalchemy as sa
from graphql import Undefined
from sqlalchemy.engine.row import Row

from ai.backend.common.data.user.types import UserRole
from ai.backend.common.types import BinarySize, ResourceSlot
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.resource_preset.types import ResourcePresetData
from ai.backend.manager.models.minilang.ordering import ColumnMapType, QueryOrderParser
from ai.backend.manager.models.minilang.queryfilter import FieldSpecType, QueryFilterParser
from ai.backend.manager.models.resource_preset import ResourcePresetRow, resource_presets
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.resource_preset.creators import ResourcePresetCreatorSpec
from ai.backend.manager.repositories.resource_preset.updaters import ResourcePresetUpdaterSpec
from ai.backend.manager.types import OptionalState, TriState

from .base import (
    BigInt,
    batch_result,
    batch_result_in_scalar_stream,
)

if TYPE_CHECKING:
    from .schema import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.models"))

__all__: Sequence[str] = (
    "CreateResourcePreset",
    "DeleteResourcePreset",
    "ModifyResourcePreset",
    "ResourcePreset",
)


class ResourcePreset(graphene.ObjectType):  # type: ignore[misc]
    id = graphene.UUID(description="Added in 25.4.0. ID of the resource preset.")
    name = graphene.String()
    resource_slots = graphene.JSONString()
    shared_memory = BigInt()
    scaling_group_name = graphene.String(
        description=(
            "Added in 25.4.0. A name of scaling group(=resource group) of the resource preset associated with."
        ),
    )

    _queryfilter_fieldspec: FieldSpecType = {
        "id": ("id", None),
        "name": ("name", None),
        "scaling_group_name": ("scaling_group_name", None),
    }

    _queryorder_colmap: ColumnMapType = {
        "id": ("id", None),
        "name": ("name", None),
        "scaling_group_name": ("scaling_group_name", None),
    }

    @classmethod
    def from_row(
        cls,
        ctx: GraphQueryContext,
        row: ResourcePresetRow | ResourcePresetData | Row[Any] | None,
    ) -> ResourcePreset | None:
        match row:
            case ResourcePresetRow():
                shared_memory = str(row.shared_memory) if row.shared_memory is not None else None
                return cls(
                    id=row.id,
                    name=row.name,
                    resource_slots=row.resource_slots.to_json(),
                    shared_memory=shared_memory,
                    scaling_group_name=row.scaling_group_name,
                )
            case ResourcePresetData():
                shared_memory = str(row.shared_memory) if row.shared_memory is not None else None
                return cls(
                    id=row.id,
                    name=row.name,
                    resource_slots=row.resource_slots.to_json(),
                    shared_memory=shared_memory,
                    scaling_group_name=row.scaling_group_name,
                )
            case Row():
                shared_memory = str(row.shared_memory) if row.shared_memory else None
                return cls(
                    id=row.id,
                    name=row.name,
                    resource_slots=row.resource_slots.to_json(),
                    shared_memory=shared_memory,
                    scaling_group_name=row.scaling_group_name,
                )
            case _:
                return None

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        *,
        filter: str | None = None,
        order: str | None = None,
    ) -> Sequence[ResourcePreset]:
        query = sa.select(ResourcePresetRow)
        if filter is not None:
            filter_parser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = filter_parser.append_filter(query, filter)
        if order is not None:
            order_parser = QueryOrderParser(cls._queryorder_colmap)
            query = order_parser.append_ordering(query, order)
        else:
            query = query.order_by(ResourcePresetRow.name)
        async with ctx.db.begin_readonly_session() as db_session:
            return [
                obj
                async for r in (await db_session.stream_scalars(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def batch_load_by_id(
        cls,
        ctx: GraphQueryContext,
        ids: Sequence[UUID],
    ) -> Sequence[ResourcePreset | None]:
        query = (
            sa.select(ResourcePresetRow)
            .where(ResourcePresetRow.id.in_(ids))
            .order_by(ResourcePresetRow.id)
        )
        async with ctx.db.begin_readonly_session() as db_session:
            return await batch_result_in_scalar_stream(
                ctx,
                db_session,
                query,
                cls,
                ids,
                lambda row: row.id,
            )

    @classmethod
    async def batch_load_by_name(
        cls,
        ctx: GraphQueryContext,
        names: Sequence[str],
    ) -> Sequence[ResourcePreset | None]:
        query = (
            sa.select(resource_presets)
            .select_from(resource_presets)
            .where(resource_presets.c.name.in_(names))
            .order_by(resource_presets.c.name)
        )
        async with ctx.db.begin_readonly() as conn:
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                names,
                lambda row: row.name,
            )


class CreateResourcePresetInput(graphene.InputObjectType):  # type: ignore[misc]
    resource_slots = graphene.JSONString(required=True)
    shared_memory = graphene.String(required=False)
    scaling_group_name = graphene.String(
        required=False,
        description=(
            "Added in 25.4.0. A name of scaling group(=resource group) of the resource preset associated with."
        ),
    )

    def to_creator(self, name: str) -> Creator[ResourcePresetRow]:
        return Creator(
            spec=ResourcePresetCreatorSpec(
                name=name,
                resource_slots=ResourceSlot.from_user_input(self.resource_slots, None),
                shared_memory=self.shared_memory if self.shared_memory else None,
                scaling_group_name=self.scaling_group_name if self.scaling_group_name else None,
            )
        )


class ModifyResourcePresetInput(graphene.InputObjectType):  # type: ignore[misc]
    name = graphene.String(
        required=False,
        description=("Added in 25.4.0. A name of resource preset."),
    )
    resource_slots = graphene.JSONString(required=False)
    shared_memory = graphene.String(required=False)
    scaling_group_name = graphene.String(
        required=False,
        description=(
            "Added in 25.4.0. A name of scaling group(=resource group) of the resource preset associated with."
        ),
    )

    def to_updater(self, id: UUID | None, name: str | None) -> Updater[ResourcePresetRow]:
        resource_slots = (
            ResourceSlot.from_json(self.resource_slots) if self.resource_slots else Undefined
        )

        pk_value = id if id is not None else (name if name is not None else "")
        return Updater(
            spec=ResourcePresetUpdaterSpec(
                resource_slots=OptionalState[ResourceSlot].from_graphql(resource_slots),
                name=OptionalState[str].from_graphql(self.name),
                shared_memory=TriState[BinarySize].from_graphql(
                    BinarySize.finite_from_str(self.shared_memory)
                    if self.shared_memory is not Undefined and self.shared_memory is not None
                    else self.shared_memory
                ),
                scaling_group_name=TriState[str].from_graphql(self.scaling_group_name),
            ),
            pk_value=pk_value,
        )


class CreateResourcePreset(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        name = graphene.String(required=True)
        props = CreateResourcePresetInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    resource_preset = graphene.Field(lambda: ResourcePreset, required=False)

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        name: str,
        props: CreateResourcePresetInput,
    ) -> CreateResourcePreset:
        from ai.backend.manager.services.resource_preset.actions.create_preset import (
            CreateResourcePresetAction,
        )

        graph_ctx: GraphQueryContext = info.context

        result = await graph_ctx.processors.resource_preset.create_preset.wait_for_complete(
            CreateResourcePresetAction(creator=props.to_creator(name))
        )

        return cls(True, "success", ResourcePreset.from_row(graph_ctx, result.resource_preset))


class ModifyResourcePreset(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        id = graphene.UUID(
            required=False,
            default_value=None,
            description=("Added in 25.4.0. ID of the resource preset."),
        )
        name = graphene.String(
            required=False, default_value=None, deprecation_reason="Deprecated since 25.4.0."
        )
        props = ModifyResourcePresetInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        id: UUID | None,
        name: str | None,
        props: ModifyResourcePresetInput,
    ) -> ModifyResourcePreset:
        from ai.backend.manager.services.resource_preset.actions.modify_preset import (
            ModifyResourcePresetAction,
        )

        graph_ctx: GraphQueryContext = info.context

        await graph_ctx.processors.resource_preset.modify_preset.wait_for_complete(
            ModifyResourcePresetAction(id=id, name=name, updater=props.to_updater(id, name))
        )

        return cls(True, "success")


class DeleteResourcePreset(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        id = graphene.UUID(
            required=False,
            default_value=None,
            description=("Added in 25.4.0. ID of the resource preset."),
        )
        name = graphene.String(
            required=False, default_value=None, deprecation_reason="Deprecated since 25.4.0."
        )

    ok = graphene.Boolean()
    msg = graphene.String()

    @classmethod
    async def mutate(
        cls,
        root: Any,
        info: graphene.ResolveInfo,
        id: UUID | None,
        name: str | None,
    ) -> DeleteResourcePreset:
        from ai.backend.manager.services.resource_preset.actions.delete_preset import (
            DeleteResourcePresetAction,
        )

        graph_ctx: GraphQueryContext = info.context

        await graph_ctx.processors.resource_preset.delete_preset.wait_for_complete(
            DeleteResourcePresetAction(id=id, name=name)
        )

        return cls(True, "success")



"""Deployment execution logic."""

import asyncio
import logging
from collections.abc import Coroutine, Mapping, Sequence
from typing import Any, cast
from uuid import UUID

from ai.backend.common.clients.http_client.client_pool import (
    ClientKey,
    ClientPool,
)
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.config import ModelHealthCheck
from ai.backend.common.exception import BackendAIError
from ai.backend.common.types import (
    RuntimeVariant,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.appproxy.client import AppProxyClient
from ai.backend.manager.clients.appproxy.types import (
    CreateEndpointRequestBody,
    EndpointTagsModel,
    SessionTagsModel,
    TagsModel,
)
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.deployment.scale import AutoScalingRule
from ai.backend.manager.data.deployment.types import (
    DeploymentInfo,
    RouteInfo,
    RouteStatus,
    RouteTrafficStatus,
)
from ai.backend.manager.data.resource.types import ScalingGroupProxyTarget
from ai.backend.manager.errors.deployment import ReplicaCountMismatch
from ai.backend.manager.errors.service import ModelDefinitionNotFound
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.repositories.base import Creator
from ai.backend.manager.repositories.base.updater import BatchUpdater
from ai.backend.manager.repositories.deployment.creators import (
    RouteBatchUpdaterSpec,
    RouteCreatorSpec,
)
from ai.backend.manager.repositories.deployment.options import RouteConditions
from ai.backend.manager.repositories.deployment.repository import (
    AutoScalingMetricsData,
    DeploymentRepository,
)
from ai.backend.manager.sokovan.deployment.definition_generator.registry import (
    ModelDefinitionGeneratorRegistry,
    RegistryArgs,
)
from ai.backend.manager.sokovan.deployment.recorder.context import DeploymentRecorderContext
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController

from .types import (
    DeploymentExecutionError,
    DeploymentExecutionResult,
)

log = BraceStyleAdapter(logging.getLogger(__name__))

REGISTER_ENDPOINT_TIMEOUT_SEC = 30


def _extract_error_code(exception: BaseException) -> str | None:
    """Extract error code from exception if available.

    Args:
        exception: The exception to extract error code from.

    Returns:
        Error code string if exception is BackendAIError, None otherwise.
    """
    if isinstance(exception, BackendAIError):
        return str(exception.error_code())
    return None


class DeploymentExecutor:
    """Executor for deployment operations."""

    _deployment_repo: DeploymentRepository
    _scheduling_controller: SchedulingController
    _config_provider: ManagerConfigProvider
    _client_pool: ClientPool
    _valkey_stat: ValkeyStatClient

    def __init__(
        self,
        deployment_repo: DeploymentRepository,
        scheduling_controller: SchedulingController,
        config_provider: ManagerConfigProvider,
        client_pool: ClientPool,
        valkey_stat: ValkeyStatClient,
    ) -> None:
        """Initialize the deployment executor."""
        self._deployment_repo = deployment_repo
        self._scheduling_controller = scheduling_controller
        self._config_provider = config_provider
        self._client_pool = client_pool
        self._valkey_stat = valkey_stat
        self._model_definition_generator_registry = ModelDefinitionGeneratorRegistry(
            RegistryArgs(
                deployment_repository=self._deployment_repo,
                enable_model_definition_override=self._config_provider.config.deployment.enable_model_definition_override,
            )
        )

    async def check_pending_deployments(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        # Phase 1: Load configuration
        with DeploymentRecorderContext.shared_phase("load_configuration"):
            with DeploymentRecorderContext.shared_step("load_proxy_targets"):
                scaling_groups = {deployment.metadata.resource_group for deployment in deployments}
                scaling_group_targets = (
                    await self._deployment_repo.fetch_scaling_group_proxy_targets(scaling_groups)
                )

        # Collect registration tasks
        registration_tasks: list[Coroutine[Any, Any, str]] = []
        valid_deployments: list[DeploymentInfo] = []
        for deployment in deployments:
            target_revision = deployment.target_revision()
            if not target_revision:
                log.warning(
                    "Deployment {} has no target revision, skipping",
                    deployment.id,
                )
                continue
            targets = scaling_group_targets[deployment.metadata.resource_group]
            if not targets:
                log.warning(
                    "No proxy target found for scaling group {}, skipping deployment {}",
                    deployment.metadata.resource_group,
                    deployment.id,
                )
                continue
            registration_tasks.append(self._register_endpoint(deployment, targets))
            valid_deployments.append(deployment)

        # Wait for all tasks to complete
        successful_deployments: list[DeploymentInfo] = []
        errors: list[DeploymentExecutionError] = []
        url_updates: dict[UUID, str] = {}

        # Phase 2: Register endpoints (per-deployment phase/step in _register_endpoint)
        if registration_tasks:
            results = await asyncio.gather(*registration_tasks, return_exceptions=True)

            for deployment, result in zip(valid_deployments, results, strict=True):
                if isinstance(result, BaseException):
                    log.error(
                        "Failed to register endpoint for deployment {}: {}",
                        deployment.id,
                        result,
                    )
                    errors.append(
                        DeploymentExecutionError(
                            deployment_info=deployment,
                            reason=str(result),
                            error_detail="Failed to register endpoint",
                            error_code=_extract_error_code(result),
                        )
                    )
                else:
                    # Result is the endpoint URL string returned from _register_endpoint
                    url_updates[deployment.id] = result
                    successful_deployments.append(deployment)
                    log.info(
                        "Successfully registered endpoint for deployment {} with URL: {}",
                        deployment.id,
                        result,
                    )

        # Phase 3: Update endpoint URLs (only for successful deployments)
        if url_updates:
            with DeploymentRecorderContext.shared_phase(
                "update_endpoint_urls", entity_ids=set(url_updates.keys())
            ):
                with DeploymentRecorderContext.shared_step("sync_endpoint_url"):
                    await self._deployment_repo.update_endpoint_urls_bulk(url_updates)

        return DeploymentExecutionResult(
            successes=successful_deployments,
            errors=errors,
        )

    async def check_ready_deployments_that_need_scaling(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        # Phase 1: Load routes
        with DeploymentRecorderContext.shared_phase("load_routes"):
            with DeploymentRecorderContext.shared_step("load_active_routes"):
                endpoint_ids = {deployment.id for deployment in deployments}
                route_map = await self._deployment_repo.fetch_active_routes_by_endpoint_ids(
                    endpoint_ids
                )

        successes: list[DeploymentInfo] = []
        errors: list[DeploymentExecutionError] = []

        # Phase 2: Verify replicas (per-deployment)
        for deployment in deployments:
            try:
                self._verify_deployment_replicas(deployment, route_map)
                successes.append(deployment)
            except ReplicaCountMismatch as e:
                log.warning(
                    "Deployment {} has mismatched active routes: {}",
                    deployment.id,
                    e,
                )
                errors.append(
                    DeploymentExecutionError(
                        deployment_info=deployment,
                        reason="Mismatched active routes",
                        error_detail=str(e),
                        error_code=_extract_error_code(e),
                    )
                )

        return DeploymentExecutionResult(
            successes=successes,
            errors=errors,
        )

    async def scale_deployment(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        # Phase 1: Load routes
        with DeploymentRecorderContext.shared_phase("load_routes"):
            with DeploymentRecorderContext.shared_step("load_active_routes"):
                endpoint_ids = {deployment.id for deployment in deployments}
                route_map = await self._deployment_repo.fetch_active_routes_by_endpoint_ids(
                    endpoint_ids
                )

        scale_out_creators: list[Creator[RoutingRow]] = []
        scale_in_route_ids: list[UUID] = []
        successes: list[DeploymentInfo] = []
        skipped: list[DeploymentInfo] = []
        errors: list[DeploymentExecutionError] = []

        # Phase 2: Evaluate scaling (per-deployment)
        for deployment in deployments:
            try:
                out_creators, in_route_ids = self._evaluate_deployment_scaling(
                    deployment, route_map
                )
                if out_creators or in_route_ids:
                    scale_out_creators.extend(out_creators)
                    scale_in_route_ids.extend(in_route_ids)
                    successes.append(deployment)
                else:
                    # No scaling action needed
                    skipped.append(deployment)
            except Exception as e:
                log.warning("Failed to scale deployment {}: {}", deployment.id, e)
                errors.append(
                    DeploymentExecutionError(
                        deployment_info=deployment,
                        reason=str(e),
                        error_detail="Failed to scale deployment",
                        error_code=_extract_error_code(e),
                    )
                )

        # Build BatchUpdater for scale in
        scale_in_updater: BatchUpdater[RoutingRow] | None = None
        if scale_in_route_ids:
            scale_in_updater = BatchUpdater(
                spec=RouteBatchUpdaterSpec(
                    status=RouteStatus.TERMINATING,
                    traffic_ratio=0.0,
                    traffic_status=RouteTrafficStatus.INACTIVE,
                ),
                conditions=[RouteConditions.by_ids(scale_in_route_ids)],
            )

        # Phase 3: Apply scaling (only for successful deployments)
        if scale_out_creators or scale_in_updater:
            with DeploymentRecorderContext.shared_phase(
                "apply_scaling", entity_ids={d.id for d in successes}
            ):
                with DeploymentRecorderContext.shared_step("scale_routes"):
                    await self._deployment_repo.scale_routes(scale_out_creators, scale_in_updater)

        return DeploymentExecutionResult(
            successes=successes,
            skipped=skipped,
            errors=errors,
        )

    async def calculate_desired_replicas(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        # Phase 1: Load autoscaling configuration
        with DeploymentRecorderContext.shared_phase("load_autoscaling_config"):
            with DeploymentRecorderContext.shared_step("load_autoscaling_rules"):
                endpoint_ids = {deployment.id for deployment in deployments}
                auto_scaling_rules = (
                    await self._deployment_repo.fetch_auto_scaling_rules_by_endpoint_ids(
                        endpoint_ids
                    )
                )

            with DeploymentRecorderContext.shared_step("load_metrics"):
                # Fetch all metrics data upfront
                metrics_data = await self._deployment_repo.fetch_metrics_for_autoscaling(
                    deployments, auto_scaling_rules
                )

        successes: list[DeploymentInfo] = []
        skipped: list[DeploymentInfo] = []
        errors: list[DeploymentExecutionError] = []
        desired_replicas_map: dict[UUID, int] = {}

        # Phase 2: Calculate replicas (per-deployment via asyncio.gather)
        calculation_tasks = [
            self._calculate_deployment_replicas(deployment, auto_scaling_rules, metrics_data)
            for deployment in deployments
        ]
        results = await asyncio.gather(*calculation_tasks, return_exceptions=True)

        for deployment, result in zip(deployments, results, strict=True):
            if isinstance(result, BaseException):
                log.warning(
                    "Failed to calculate desired replicas for deployment {}: {}",
                    deployment.id,
                    result,
                )
                errors.append(
                    DeploymentExecutionError(
                        deployment_info=deployment,
                        reason=str(result),
                        error_detail="Failed to calculate desired replicas",
                        error_code=_extract_error_code(result),
                    )
                )
            elif result is None:
                skipped.append(deployment)
            else:
                desired_replicas_map[deployment.id] = result
                successes.append(deployment)

        # Phase 3: Save scaling decision (only for successful deployments)
        if desired_replicas_map:
            with DeploymentRecorderContext.shared_phase(
                "save_scaling_decision", entity_ids=set(desired_replicas_map.keys())
            ):
                with DeploymentRecorderContext.shared_step("save_target_replicas"):
                    await self._deployment_repo.update_desired_replicas_bulk(desired_replicas_map)

        return DeploymentExecutionResult(
            successes=successes,
            skipped=skipped,
            errors=errors,
        )

    async def destroy_deployment(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        # Phase 1: Load termination configuration
        with DeploymentRecorderContext.shared_phase("load_termination_config"):
            with DeploymentRecorderContext.shared_step("load_routes"):
                endpoint_ids = {deployment.id for deployment in deployments}
                routes = await self._deployment_repo.fetch_active_routes_by_endpoint_ids(
                    endpoint_ids
                )

            with DeploymentRecorderContext.shared_step("load_proxy_config"):
                scaling_groups = {deployment.metadata.resource_group for deployment in deployments}
                proxy_targets = await self._deployment_repo.fetch_scaling_group_proxy_targets(
                    scaling_groups
                )

        route_ids: set[UUID] = set()
        for route_list in routes.values():
            for route in route_list:
                route_ids.add(route.route_id)

        # Phase 2: Terminate routes
        with DeploymentRecorderContext.shared_phase("terminate_routes"):
            with DeploymentRecorderContext.shared_step("mark_routes_terminating"):
                await self._deployment_repo.mark_terminating_route_status_bulk(route_ids)

        successes: list[DeploymentInfo] = []
        errors: list[DeploymentExecutionError] = []

        # Phase 3: Unregister endpoints (per-deployment via asyncio.gather)
        unregister_tasks = [
            self._unregister_endpoint(deployment, proxy_targets) for deployment in deployments
        ]
        results = await asyncio.gather(*unregister_tasks, return_exceptions=True)

        for deployment, result in zip(deployments, results, strict=True):
            if isinstance(result, BaseException):
                log.warning(
                    "Failed to unregister endpoint {}: {}",
                    deployment.id,
                    result,
                )
                errors.append(
                    DeploymentExecutionError(
                        deployment_info=deployment,
                        reason="Failed to unregister endpoint",
                        error_detail=str(result),
                        error_code=_extract_error_code(result),
                    )
                )
            else:
                successes.append(deployment)

        return DeploymentExecutionResult(
            successes=successes,
            errors=errors,
        )

    # Private helper methods

    async def _register_endpoint(
        self,
        deployment: DeploymentInfo,
        scaling_group_target: ScalingGroupProxyTarget,
    ) -> str:
        pool = DeploymentRecorderContext.current_pool()
        recorder = pool.recorder(deployment.id)

        with recorder.phase("register_endpoint"):
            with recorder.step("check_target_revision"):
                target_revision = deployment.target_revision()
                if not target_revision:
                    raise ModelDefinitionNotFound(
                        f"No target revision for deployment {deployment.id}"
                    )

            with recorder.step("generate_model_definition"):
                model_definition = (
                    await self._model_definition_generator_registry.generate_model_definition(
                        target_revision
                    )
                )
                health_check_config = model_definition.health_check_config()
                if not health_check_config:
                    log.debug(
                        "No health check configuration found in model definition for deployment {}",
                        deployment.id,
                    )

            with recorder.step("register_to_proxy"):
                return await self._create_endpoint_in_proxy(
                    endpoint_id=deployment.id,
                    endpoint_name=deployment.metadata.name,
                    session_owner_id=deployment.metadata.session_owner,
                    project_id=deployment.metadata.project,
                    domain_name=deployment.metadata.domain,
                    runtime_variant=target_revision.execution.runtime_variant,
                    existing_url=deployment.network.url,
                    open_to_public=deployment.network.open_to_public,
                    health_check_config=health_check_config,
                    app_proxy_addr=scaling_group_target.addr,
                    app_proxy_api_token=scaling_group_target.api_token,
                )

    def _load_app_proxy_client(self, address: str, token: str) -> AppProxyClient:
        """Load or create a App Proxy client for the given address."""
        client_session = self._client_pool.load_client_session(
            ClientKey(
                endpoint=address,
                domain="wsproxy",
            )
        )
        return AppProxyClient(client_session, address, token)

    async def _create_endpoint_in_proxy(
        self,
        endpoint_id: UUID,
        endpoint_name: str,
        session_owner_id: UUID,
        project_id: UUID,
        domain_name: str,
        runtime_variant: RuntimeVariant,
        existing_url: str | None,
        open_to_public: bool,
        health_check_config: ModelHealthCheck | None,
        app_proxy_addr: str,
        app_proxy_api_token: str,
    ) -> str:
        """
        Create an endpoint in WSProxy service.

        Args:
            endpoint_id: Endpoint UUID
            endpoint_name: Name of the endpoint
            session_owner_id: UUID of the session owner
            project_id: UUID of the project
            domain_name: Domain name
            runtime_variant: Runtime variant
            existing_url: Existing URL if any
            open_to_public: Public accessibility flag
            health_check_config: Optional health check configuration
            wsproxy_addr: WSProxy service address
            wsproxy_api_token: WSProxy API token

        Returns:
            Response from WSProxy service
        """
        app_proxy_client = self._load_app_proxy_client(app_proxy_addr, app_proxy_api_token)

        # Create request body using Pydantic model
        request_body = CreateEndpointRequestBody(
            version="v2",
            service_name=endpoint_name,
            tags=TagsModel(
                session=SessionTagsModel(
                    user_uuid=str(session_owner_id),
                    group_id=str(project_id),
                    domain_name=domain_name,
                ),
                endpoint=EndpointTagsModel(
                    id=str(endpoint_id),
                    runtime_variant=runtime_variant.value,
                    existing_url=str(existing_url) if existing_url else None,
                ),
            ),
            apps={},
            open_to_public=open_to_public,
            health_check=health_check_config,
        )

        res = await app_proxy_client.create_endpoint(endpoint_id, request_body)
        return cast(str, res["endpoint"])

    async def _delete_endpoint_from_wsproxy(
        self,
        endpoint_id: UUID,
        app_proxy_addr: str,
        app_proxy_api_token: str,
    ) -> None:
        """
        Delete an endpoint from WSProxy service.

        Args:
            endpoint_id: Endpoint UUID to delete
            wsproxy_addr: WSProxy service address
            wsproxy_api_token: WSProxy API token
        """
        app_proxy_client = self._load_app_proxy_client(app_proxy_addr, app_proxy_api_token)
        await app_proxy_client.delete_endpoint(endpoint_id)

    def _verify_deployment_replicas(
        self,
        deployment: DeploymentInfo,
        route_map: Mapping[UUID, Sequence[RouteInfo]],
    ) -> None:
        """Verify that deployment has the expected number of active routes."""
        pool = DeploymentRecorderContext.current_pool()
        recorder = pool.recorder(deployment.id)
        with recorder.phase("verify_replicas"):
            with recorder.step("compare_route_count"):
                routes = route_map[deployment.id]
                if len(routes) != deployment.replica_spec.target_replica_count:
                    raise ReplicaCountMismatch(
                        expected=deployment.replica_spec.target_replica_count,
                        actual=len(routes),
                    )

    def _evaluate_deployment_scaling(
        self,
        deployment: DeploymentInfo,
        route_map: Mapping[UUID, Sequence[RouteInfo]],
    ) -> tuple[list[Creator[RoutingRow]], list[UUID]]:
        """Evaluate scaling action for a deployment and return creators/route IDs."""
        pool = DeploymentRecorderContext.current_pool()
        recorder = pool.recorder(deployment.id)

        scale_out_creators: list[Creator[RoutingRow]] = []
        scale_in_route_ids: list[UUID] = []

        with recorder.phase("evaluate_scaling"):
            with recorder.step("calculate_scale_action"):
                target_count = deployment.replica_spec.target_replica_count
                routes = route_map[deployment.id]
                if len(routes) < target_count:
                    # Build creators for scale out
                    new_replica_count = target_count - len(routes)
                    for _ in range(new_replica_count):
                        creator_spec = RouteCreatorSpec(
                            endpoint_id=deployment.id,
                            session_owner_id=deployment.metadata.session_owner,
                            domain=deployment.metadata.domain,
                            project_id=deployment.metadata.project,
                            revision_id=deployment.current_revision_id,
                        )
                        scale_out_creators.append(Creator(spec=creator_spec))
                elif len(routes) > target_count:
                    termination_route_candidates = sorted(
                        routes, key=lambda r: (r.status.termination_priority())
                    )
                    candidates = termination_route_candidates[: len(routes) - target_count]
                    scale_in_route_ids.extend(r.route_id for r in candidates)

        return scale_out_creators, scale_in_route_ids

    async def _calculate_deployment_replicas(
        self,
        deployment: DeploymentInfo,
        auto_scaling_rules: Mapping[UUID, Sequence[AutoScalingRule]],
        metrics_data: AutoScalingMetricsData,
    ) -> int | None:
        """Calculate desired replicas for a single deployment.

        Returns:
            int: new desired replica count
            None: no change needed (skip)
            Raises exception: on failure
        """
        pool = DeploymentRecorderContext.current_pool()
        recorder = pool.recorder(deployment.id)

        with recorder.phase("calculate_replicas"):
            auto_scaling_rule = auto_scaling_rules.get(deployment.id, [])

            if not auto_scaling_rule:
                with recorder.step("apply_manual_scaling"):
                    routes = metrics_data.routes_by_endpoint.get(deployment.id, [])
                    if deployment.replica_spec.replica_count != len(routes):
                        return deployment.replica_spec.replica_count
                    return None

            with recorder.step("evaluate_autoscaling_rules"):
                desired_replica = (
                    await self._deployment_repo.calculate_desired_replicas_for_deployment(
                        deployment,
                        auto_scaling_rule,
                        metrics_data,
                    )
                )

                if desired_replica is None:
                    log.debug(
                        "No change in desired replicas for deployment {}, skipping",
                        deployment.id,
                    )
                return desired_replica

    async def _unregister_endpoint(
        self,
        deployment: DeploymentInfo,
        proxy_targets: Mapping[str, ScalingGroupProxyTarget | None],
    ) -> None:
        """Unregister an endpoint from the proxy service."""
        pool = DeploymentRecorderContext.current_pool()
        recorder = pool.recorder(deployment.id)

        with recorder.phase("unregister_endpoint"):
            with recorder.step("delete_from_proxy"):
                target = proxy_targets.get(deployment.metadata.resource_group)
                if not target:
                    log.warning(
                        "No proxy target found for scaling group {}, skipping unregister for {}",
                        deployment.metadata.resource_group,
                        deployment.id,
                    )
                    return
                await self._delete_endpoint_from_wsproxy(
                    endpoint_id=deployment.id,
                    app_proxy_addr=target.addr,
                    app_proxy_api_token=target.api_token,
                )



"""
Data types for domain service.
Deprecated: use `ai.backend.manager.data.domain.types` instead.
"""

from ai.backend.manager.data.domain.types import (
    DomainData,
    DomainModifier,
    DomainNodeModifier,
    UserInfo,
)
from ai.backend.manager.types import OptionalState, PartialModifier, TriState

__all__ = [
    "DomainData",
    "DomainModifier",
    "DomainNodeModifier",
    "OptionalState",
    "PartialModifier",
    "TriState",
    "UserInfo",
]
